#include <Python.h>
#include <rivet.h>

/* "Pyro" is an alternate name that won't collide with rivet names.
 * It expands to PYthon interface to Rivet Objects.
 */

#define PYRO_UNIQUE_TOKEN 0xabc /* for bindings */
static char *not_inited = "Rivet not initialized";
static char *no_rivetobj = "Object has no associated rivetobj";
static PyObject *pyro_error;
static int pyro_call_callback2(Rivetobj, XEvent*, ClientData, ClientData, int);

/*----------------------------------------------------------------------*/
/* These data structures and macros just wrap PyMethodDef structures
 * in case I want to use something better than Py_FindMethod routine, say
 * that used a binary search.
 */
typedef PyMethodDef Pyrometh;

#define PYROMDECL(_name_, _func_)	\
	{_name_, (PyCFunction)_func_, METH_VARARGS|METH_KEYWORDS}

#define PYROMDECL2(_name_, _func_)	\
	{ _name_, (PyCFunction)_func_, METH_VARARGS}

#define pyro_init_methods( methodlist ) methodlist
#define pyro_find_method(_methodlist, _obj, _name)	\
	Py_FindMethod((PyMethodDef*)_methodlist, (PyObject*)_obj, _name)

/*----------------------------------------------------------------------*/
typedef struct {
	PyObject_HEAD
	PyObject *name;
	int (*func)();
	Pyrometh *methods;
	ClientData data;
} Pyro_image_struct;
typedef Pyro_image_struct *Pyroimage;

staticforward PyTypeObject PyroImageType;
#define is_pyroimage(_A) ((_A)->ob_type == &PyroImageType)
/*----------------------------------------------------------------------*/
/*
 * "method sets" are collections of tk functions that need to be visible
 * but in their own namespaces.  Often these are tk methods not
 * associated with a widget, such as pack, winfo, etc.
 */
typedef struct {
	PyObject_HEAD
	Pyrometh *methods;
	PyObject *(*callfn)();
} Pyro_method_set_struct;
typedef Pyro_method_set_struct *Pyro_ms;

staticforward PyTypeObject PyroMsType;
#define is_pyromethodset(_A) ((_A)->ob_type == &PyroMsType)

static PyObject *
pyro_ms_getattr(PyObject *obj, char *name)
{
	Pyro_ms ms;
	PyObject *rv, *rv0;

	if(!rivet_main_window) {
		PyErr_SetString(pyro_error, not_inited);
		return NULL;
	}
	if(!obj || !is_pyromethodset(obj)) {
		PyErr_SetString(pyro_error, "No object");
		return NULL;
	}
	ms = (Pyro_ms) obj;
	return pyro_find_method(ms->methods, obj, name);
}

static PyObject *
pyro_ms_call(PyObject *obj, PyObject *args, PyObject *d)
{
	Pyro_ms ms;

	if(!obj || !is_pyromethodset(obj)) {
		PyErr_SetString(pyro_error, "No object");
		return NULL;
	}
	ms = (Pyro_ms) obj;
	if(!ms->callfn) {
		PyErr_SetString(PyExc_TypeError, "call of non-function");
		return NULL;
	}
	return (ms->callfn)(obj, args, d);
}

static PyTypeObject PyroMsType = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,				/* size */
	"rivet command",		/* name */
	sizeof(Pyro_method_set_struct),	/* size */
	0,				/* itemsize */
	(destructor)0,		 	/* dealloc */
	0,				/* print */
	(getattrfunc)pyro_ms_getattr,	/* getattr */
	(setattrfunc)0,			/* setattr */
	0,				/* compare */
	0,				/* repr */
	0,				/* as number */
	0,				/* as sequence */
	0, 				/* as mapping */
 	0,				/* tp_hash */
 	(ternaryfunc)pyro_ms_call,	/* tp_call */
 	0,				/* tp_str */
};

/*----------------------------------------------------------------------*/

typedef struct {
	PyObject_HEAD
	Rivetobj rivetobj;
	PyObject *dictionary; /* can be NULL */
} Pyro_object_struct;
typedef Pyro_object_struct *Pyroobj;

staticforward PyTypeObject PyroType;
#define is_pyroobj(_A) ((_A)->ob_type == &PyroType)

static Pyroobj
pyro_get_widget(Pyroobj in)
{
	PyObject *obj;

	obj = (PyObject*)in;

	if(obj && PyInstance_Check(obj)) {
		obj = PyObject_GetAttrString(obj, "__widget");
		/* Must decref because this function assumes 
		 * ref count neutrality.
		 */
		Py_XDECREF(obj);
	}
	if(!obj || !is_pyroobj(obj)) {
		PyErr_SetString(pyro_error, 
		    "Object is not a rivetobj or has no associated rivetobj");
		return 0;
	}
	if(!((Pyroobj)obj)->rivetobj) {
		PyErr_SetString(pyro_error, "Tk widget has been destroyed");
		return 0;
	}
	return (Pyroobj) obj;
}

/*----------------------------------------------------------------------*/

static int
pyro_callable_check(PyObject *callbackobj, int generate_errors)
{
	PyObject *checkee;

	if(PyTuple_Check(callbackobj)) {
		if(PyTuple_Size(callbackobj) < 1) {
			if(generate_errors)
			    PyErr_SetString(pyro_error,
				"First Tuple arg should be callable");
			return 0;
		}
		checkee = PyTuple_GetItem(callbackobj, 0);
	} else
		checkee = callbackobj;

	if(PyCallable_Check(checkee))
		return 1;

	if(generate_errors)
		PyErr_SetString(pyro_error, "Arg should be callable"); 
	return 0;
}

/*----------------------------------------------------------------------*/
/* arg list processing macros and structures */

#define MAX_ARG_LEN 40
#define PYRO_MIN_ARGS 20
typedef struct _tmp_buf_list {
	struct _tmp_buf_list *next;
	char space[MAX_ARG_LEN];
} Tmpbuf;
typedef struct {
	/* for most small arglists, use space on stack. */
	char *fast_argv[PYRO_MIN_ARGS];
	char fast_tmpbuf[PYRO_MIN_ARGS][MAX_ARG_LEN];
	int tmpbufcount; /* only applies to fast_tmpbuf */

	/* for calls with arbitrarily length arglists, dynamically
	 * allocate space. */
	char **argv;
	int argv_len;	/* how much is allocated */
	int argc;	/* how much is in use */

	/* large arglist temporary buffers are allocated individually */
	Tmpbuf *tmpbuf;
} Argcv;

#define INIT_ARGCV(_argcv, _initcount) {		\
		_argcv.argc = _initcount;		\
		_argcv.argv_len = PYRO_MIN_ARGS;	\
		_argcv.tmpbufcount = 0;			\
		_argcv.tmpbuf = 0;			\
		_argcv.argv = _argcv.fast_argv;		\
		}

#define FREE_ARGCV(_argcv) 				\
		if(argcv.argc > PYRO_MIN_ARGS)	pyro_free_argcv(&_argcv);

static void
pyro_free_argcv(Argcv *argcv)
{
	Tmpbuf *tb, *tbnext;
	tb = argcv->tmpbuf;
	while(tb) {
		tbnext = tb->next;
		free(tb);
		tb = tbnext;
	}
	argcv->tmpbuf = 0;
	if(argcv->argc > PYRO_MIN_ARGS && argcv->argv) {
		free(argcv->argv);
		argcv->argv = 0;
	}
	argcv->argc = 0;
}

static int
pyro_parse_one_arg(Argcv *argcv, PyObject *arg, int hyphenate)
{
#define GROWBY 100
	Tmpbuf *tb;
	char *result;
	int type;

	again:
	if(PyString_Check(arg)) {
		if(hyphenate) 
			type = 5;
		else {
			type = 1;
			result = PyString_AsString(arg);
		}
	} else
	if(is_pyroobj(arg)) {
		type = 1;
		if(!((Pyroobj)arg)->rivetobj)
			return 0; /* look out for destroyed objects */
		result = rivet_path(((Pyroobj)arg)->rivetobj);
	} else
	if(PyInt_Check(arg)) {
		type = 2;
	} else
	if(PyInstance_Check(arg)) {
		PyObject *obj;
		obj = PyObject_GetAttrString(arg, "__widget");
		if(obj) {
			Py_DECREF(obj); /* need to be ref count neutral */
			arg = obj;
			goto again;
		} else {
			PyErr_Clear();
			/* not a widget wrapper, check for a variable name. */
			obj = PyObject_GetAttrString(arg, "__var");
			if(obj) {
				Py_DECREF(obj); /* ref count neutral */
				if(PyString_Check(obj)) {
					type = 1;
					result = PyString_AsString(obj);
				} else {
					return 0;
				}
			} else {
				PyErr_Clear();
				return 0;
			}
		}
	} else
	if(pyro_callable_check(arg, 0)) {
		type = 1;
		result = rivet_new_callback(
				pyro_call_callback2, arg, PYRO_UNIQUE_TOKEN);
	} else
	if(PyLong_Check(arg)) {
		type = 3;
	} else
	if(PyFloat_Check(arg)) {
		type = 4;
	} else 
	if(is_pyroimage(arg)) {
		type = 1;
		if(!((Pyroimage) arg)->name)
			return 0; /* look out for deleted images */
		result = PyString_AsString((PyObject*)((Pyroimage) arg)->name);
	} else
	if(arg == Py_None) {
		type = 1;
		result = "";
	} else
		return 0; /* type not supported */

	/* Store the result.  First, for things that need to
	 * be converted to strings, get some memory for the
	 * result, then store it.
	 */
	if(type > 1) {
		if(argcv->argc < PYRO_MIN_ARGS) {
			result = argcv->fast_tmpbuf[argcv->tmpbufcount];
			argcv->tmpbufcount++;
		} else {
			/* no more fast space, left, do args. */
			tb = (Tmpbuf*) malloc(sizeof(Tmpbuf));
			if(!tb) {
				pyro_free_argcv(argcv);
				return 0;
			}
			tb->next = argcv->tmpbuf;
			argcv->tmpbuf = tb;
			result = tb->space;
		}
		switch(type) {
		  case 2:
			sprintf(result, "%ld", PyInt_AsLong(arg));
			break;
		  case 3:
			sprintf(result, "%ld", PyLong_AsLong(arg));
			break;
		  case 4:
			sprintf(result, "%g", PyFloat_AsDouble(arg));
			break;
		  case 5: /* hyphenate a string */
			result[0] = '-';
			strncpy(result+1, PyString_AsString(arg),
				MAX_ARG_LEN-2);
			break;
		}
	}

	/* Put the result in the arglist. */
	if(argcv->argc < PYRO_MIN_ARGS) {
		argcv->argv[argcv->argc] = result;
		argcv->argc++;
		return 1;
	}

	if(argcv->argc == argcv->argv_len) {
		/* have to grow the argv, the first one is grown
		 * by malloc and copy, the remainder by realloc. */
		char **newargv;
		argcv->argv_len = argcv->argc+GROWBY;
		if(argcv->argc == PYRO_MIN_ARGS) {
			newargv = (char**)malloc(argcv->argv_len*sizeof(char**));
			if(!newargv) {
				pyro_free_argcv(argcv);
				return 0;
			}
			memcpy(newargv, argcv->argv, 
				PYRO_MIN_ARGS*sizeof(char**));
		} else {
			newargv = (char**)realloc(argcv->argv, 
					argcv->argv_len*sizeof(char**));
			if(!newargv) {
				pyro_free_argcv(argcv);
				return 0;
			}
		}
		argcv->argv = newargv;
	}

	argcv->argv[argcv->argc] = result;
	argcv->argc++;
	return 1;
#undef GROWBY
}

static int
pyro_parse_argcv(Argcv *argcv, PyObject *args_in)
{
	int len;
	int i;
	PyObject *args;
	PyObject *arg;
	PyObject *(*getfunction)(PyObject*, int);

	/* Flatten tuple and list args into an arglist.
	 * argcv->argc should be previously set to skip over space already
	 * set by caller, for both recursion and prefix args.
	 */

	if(PyTuple_Check(args_in)) {
		args = args_in;
		len = PyTuple_Size(args);
		getfunction = PyTuple_GetItem;
	} else
	if(PyList_Check(args_in)) {
		args = args_in;
		len = PyList_Size(args);
		getfunction = PyList_GetItem;
	} else
	if(PyDict_Check(args_in)) {
		/* Convert a dictionary to a list of tuples, */
		/* then parse the first item of each tuple with hyphenation. */
		
		args = PyDict_Items(args_in);
		if(!args)
			return 1;
		len = PyList_Size(args);
		for(i=0; i<len; i++) {
		    PyObject *kw, *val;

		    arg = PyList_GetItem(args, i); /* returns a tuple */
		    kw = PyTuple_GetItem(arg, 0);
		    val = PyTuple_GetItem(arg, 1);

		    if (pyro_parse_one_arg(argcv, kw, 1) == 0) {
			PyErr_SetString(pyro_error, "Bad keyword.");
			Py_DECREF(args);
			return 1;
		    }
		    if (pyro_parse_one_arg(argcv, val, 0) == 0) {
			PyErr_SetString(pyro_error, "Bad keyword arg value.");
			Py_DECREF(args);
			return 1;
		    }
		}
		Py_DECREF(args); /* clean up temporary */
		return 0;
	} else {
		if(pyro_parse_one_arg(argcv, args_in, 0))
			return 0;
		PyErr_SetString(pyro_error, "Cannot handle type.");
		pyro_free_argcv(argcv);
		return 1;
	}

	for(i=0; i<len; i++) {
		arg = (getfunction)(args, i);

		if(pyro_parse_one_arg(argcv, arg, 0))
			continue;

		if(PyTuple_Check(arg) || PyList_Check(arg) ||
		   PyDict_Check(arg)) {
			/* call recursively */
			if(pyro_parse_argcv(argcv, arg))
				return 1;
			continue;
		}

		/* The arg at index argcv->argc is no good. */
		PyErr_SetString(pyro_error, "Bad arg.");
		pyro_free_argcv(argcv);
		if(PyDict_Check(args_in))
			Py_DECREF(args);
		return 1;
	}
	return 0;
}

/*----------------------------------------------------------------------*/

static PyObject *
pyro_destroy_object(PyObject *self, PyObject *args)
{
	Pyroobj po;

	if(self) {
		po = pyro_get_widget((Pyroobj)self);
	} else {
		if(!PyArg_ParseTuple(args, "O", &po))
			return NULL;
		po = pyro_get_widget(po);
	}
	if(!po)
		return NULL;
	rivet_destroy_window(po->rivetobj);
	Py_INCREF(Py_None);
	return Py_None;
}

static char *
pyro_get_childname(PyObject *args, int index)
{
	/* If child is named at index in tuple, return the string for it */
	PyObject *obj;
	char *name;

	obj = PyTuple_GetItem(args, index);
	if(!obj) {
		PyErr_Clear(); /* might be out of range */
		return 0;
	}
	if(PyString_Check(obj)) {
		name = PyString_AsString(obj);
		if(name && name[0] == '.')
			return name;
	}
	return 0;
}

static PyObject *
pyro_create_object(Rivetclass rclass, Pyrometh *classmethods,
	Pyroobj parent_arg, PyObject *args, PyObject *d)
{
	/* rclass == NULL indicates create top level */
	Pyroobj po;
	Rivetobj ro;
	Argcv argcv;
	Pyroobj parent;
	char *childname;

	if(parent_arg)
		parent = pyro_get_widget(parent_arg); /* supplied by self */
	else {
		if(!rivet_main_window) {
			PyErr_SetString(pyro_error, not_inited);
			return NULL;
		}
		/* a function call, parent should be first arg */
		parent = (Pyroobj) PyTuple_GetItem(args, 0);
		parent = pyro_get_widget(parent);
	}

	if(!parent) {
		PyErr_SetString(pyro_error, "Illegal parent widget.");
		return NULL;
	}

	if(!rivet_is_class_initialized(rclass))
		rclass->language_class_object = (ClientData) classmethods;

	po = PyObject_NEW(Pyro_object_struct, &PyroType);
	if(!po) {
		PyErr_SetString(pyro_error, "Could not create rivet object.");
		return NULL;
	}
	po->rivetobj = 0;
	po->dictionary = 0;

	INIT_ARGCV(argcv, 0)
	/* Here we check a few things to get argv lined up so that
	 * argv[2] starts the list of arguments, regardless of specification
	 * of parent or name in the tuple argument.  This is probably 
	 * complicated by the optional name. 
	 */
	if(parent_arg) {
		/* parent is not part of tuple arg, it was "self" */
		if(childname = pyro_get_childname(args, 0))
			argcv.argc = 1; /*eg: parent.widget(name, args...) */
		else
			argcv.argc = 2; /*eg: parent.widget(args...) */
	} else {
		/* parent is part of tuple arg */
		if(childname = pyro_get_childname(args, 1))
			argcv.argc = 0; /*eg: rivet.widget(parent, name, a...)*/
		else 
			argcv.argc = 1; /* eg: rivet.widget(parent, args...) */
	}

	if(pyro_parse_argcv(&argcv, args)) {
		Py_DECREF(po);
		return NULL;
	}
	if(d && pyro_parse_argcv(&argcv, d)) {
		Py_DECREF(po);
		return NULL;
	}
	argcv.argv[1] = childname;

	ro = rivet_create(rclass, parent->rivetobj, argcv.argc, argcv.argv);
	FREE_ARGCV(argcv)

	if(!ro) {
		PyErr_SetString(pyro_error, parent->rivetobj->interp->result);
		Py_DECREF(po);
		return NULL;
	}

	po->rivetobj = ro;
	ro->language_object = (ClientData) po;
	Py_INCREF(po); /* add a reference from the parent to child */

	return (PyObject*) po;
}

static Pyrometh *simple_widget_methods;
static Pyrometh *toplevel_widget_methods;
static Pyrometh *button_widget_methods;
static Pyrometh *cbutton_widget_methods;
static Pyrometh *rbutton_widget_methods;
static Pyrometh *entry_widget_methods;
static Pyrometh *scale_widget_methods;
static Pyrometh *canvas_widget_methods;
static Pyrometh *listbox_widget_methods;
static Pyrometh *menu_widget_methods;
static Pyrometh *scrollbar_widget_methods;
static Pyrometh *text_widget_methods;


#define NEWWIDGETMETH( _methodname, _rivetclass, _classmethods )	\
	static PyObject *						\
	_methodname (PyObject *parent, PyObject *args, PyObject *d)	\
	{								\
	  return pyro_create_object( _rivetclass, _classmethods,	\
		(Pyroobj) parent, args, d);				\
	}
NEWWIDGETMETH(pyro_button,	ButtonClass, button_widget_methods)
NEWWIDGETMETH(pyro_canvas,	CanvasClass, canvas_widget_methods)
NEWWIDGETMETH(pyro_checkbutton,	CheckButtonClass, cbutton_widget_methods)
NEWWIDGETMETH(pyro_entry,	EntryClass, entry_widget_methods)
NEWWIDGETMETH(pyro_frame,	FrameClass, simple_widget_methods)
NEWWIDGETMETH(pyro_label,	LabelClass, simple_widget_methods)
NEWWIDGETMETH(pyro_listbox,	ListboxClass, listbox_widget_methods)
NEWWIDGETMETH(pyro_menu,	MenuClass, menu_widget_methods)
NEWWIDGETMETH(pyro_menubutton,	MenubuttonClass, simple_widget_methods)
NEWWIDGETMETH(pyro_message,	MessageClass, simple_widget_methods)
NEWWIDGETMETH(pyro_radiobutton,	RadioButtonClass, rbutton_widget_methods)
NEWWIDGETMETH(pyro_scale,	ScaleClass, scale_widget_methods)
NEWWIDGETMETH(pyro_scrollbar,	ScrollbarClass, scrollbar_widget_methods)
NEWWIDGETMETH(pyro_text,	TextClass, text_widget_methods)
NEWWIDGETMETH(pyro_toplevel,	ToplevelClass, toplevel_widget_methods)
#undef NEWWIDGETMETH

/*----------------------------------------------------------------------*/

/* converter from tk return values to python types */

static PyObject *
pyro_conv_int(char *str, PyObject *args)
{
	long x;
	x = atol(str);
	return PyInt_FromLong(x);
}

static PyObject *
pyro_conv_int_none(char *str, PyObject *args)
{
	long x;
	if(str) {
		x = atol(str);
		return PyInt_FromLong(x);
	} 
	Py_INCREF(Py_None);
	return Py_None;
}


static PyObject *
pyro_conv_float(char *str, PyObject *args)
{
	double x;
	x = strtod(str, 0);
	return PyFloat_FromDouble(x);
}

static PyObject *
pyro_conv_none(char *str, PyObject *args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
pyro_conv_str(char *str, PyObject *args)
{
	if(str && str[0] != 0)
		return PyString_FromString(str);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
pyro_conv_str2(char *str, PyObject *args)
{
	/* this one preserves null strings */
	return PyString_FromString(str);
}

static PyObject *
pyro_conv_rivetobj(char *str_in, PyObject *args)
{
	/* convert .splat.splat pattern to a single pyroobj, or none */
	PyObject *rv;
	Rivetobj ro;

	ro = rivet_path_to_rivetobj(str_in);
	if(ro && ro->language_object) {
		rv = (PyObject*) ro->language_object;
	} else {
		rv = Py_None;
	}
	Py_INCREF(rv);
	return rv;
}

static PyObject *
pyro_conv_tuple_str(char *str_in, PyObject *args)
{
	int i;
	int argc;
	char **argv;
	PyObject *rv;
	PyObject *str;

	/* rip apart things assembled by Tcl_AppendResult. */

	rv = 0;
	if(Tcl_SplitList(rivet_main_window->interp, str_in,
				&argc, &argv) != TCL_ERROR) {
		rv = PyTuple_New(argc);
		for(i=0;i<argc;i++) {
			str = PyString_FromString(argv[i]);
	    		PyTuple_SetItem(rv, i, str);
		}
		free((char*) argv);
	}
	if(!rv) {
		Py_INCREF(Py_None);
		return Py_None;
	}
	return rv;
}

static PyObject *
pyro_conv_str_tuple_str(char *str_in, PyObject *args)
{
	PyObject *rv;
	if(PyTuple_Size(args) == 1)
		return pyro_conv_str2(str_in, args);
	return pyro_conv_tuple_str(str_in, args);
}

static PyObject *
pyro_conv_tuple_int(char *str_in, PyObject *args)
{
	int i;
	int argc;
	char **argv;
	PyObject *rv;
	PyObject *val;

	/* rip apart things assembled by Tcl_AppendResult. */

	rv = 0;
	if(Tcl_SplitList(rivet_main_window->interp, str_in,
				&argc, &argv) != TCL_ERROR) {
		rv = PyTuple_New(argc);
		for(i=0;i<argc;i++) {
			val = PyInt_FromLong(atol(argv[i]));
	    		PyTuple_SetItem(rv, i, val);
		}
		free((char*) argv);
	}
	if(!rv) {
		Py_INCREF(Py_None);
		return Py_None;
	}
	return rv;
}

static PyObject *
pyro_conv_tuple_float(char *str_in, PyObject *args)
{
	int i;
	int argc;
	char **argv;
	PyObject *rv;
	PyObject *val;

	/* rip apart things assembled by Tcl_AppendResult. */

	rv = 0;
	if(Tcl_SplitList(rivet_main_window->interp, str_in,
				&argc, &argv) != TCL_ERROR) {
		rv = PyTuple_New(argc);
		for(i=0;i<argc;i++) {
			val = PyFloat_FromDouble(strtod(argv[i], 0));
	    		PyTuple_SetItem(rv, i, val);
		}
		free((char*) argv);
	}
	if(!rv) {
		Py_INCREF(Py_None);
		return Py_None;
	}
	return rv;
}

static PyObject *
pyro_conv_tuple_rivetobj(char *str_in, PyObject *args)
{
	/* convert .splat.splat patterns to pyroobjs, and rest to strings. */
	int i;
	int argc;
	char **argv;
	PyObject *rv;
	PyObject *val;
	Rivetobj ro;

	/* rip apart things assembled by Tcl_AppendResult. */

	rv = 0;
	if(Tcl_SplitList(rivet_main_window->interp, str_in,
				&argc, &argv) != TCL_ERROR) {
		rv = PyTuple_New(argc);
		for(i=0;i<argc;i++) {
			if((argv[i][0] == '.')  &&
			   (ro = rivet_path_to_rivetobj(argv[i])) &&
			    ro->language_object) {
				val = (PyObject*) ro->language_object;
				Py_INCREF(val);
			} else {
				val = PyString_FromString(argv[i]);
			}
	    		PyTuple_SetItem(rv, i, val);
		}
		free((char*) argv);
	}
	if(!rv) {
		Py_INCREF(Py_None);
		return Py_None;
	}
	return rv;
}


static PyObject *
pyro_conv_passthrough(char *str, PyObject *args)
{
	return (PyObject *) str;
}

static PyObject *
pyro_conv_cget(char *str, PyObject *args)
{
	/* see tkConfig.c: FormatConfigValue */
	switch (rivetval_cget.type) {
		case Rivet_str:
			return PyString_FromString(rivetval_cget.strval);
		case Rivet_int:
			return PyInt_FromLong(rivetval_cget.intval);
		case Rivet_double:
			return PyFloat_FromDouble(rivetval_cget.doubleval);
		case Rivet_callback: {
			Callback cb;
			cb = rivetval_cget.callbackval;
			if(cb && cb->type == PYRO_UNIQUE_TOKEN && cb->assoc) {
				Py_INCREF((PyObject*) cb->assoc);
				return (PyObject*) cb->assoc;
			}
			break;
		}
		case Rivet_rivetobj: {
			PyObject *po;
			po = (PyObject*)rivetval_cget.rivetobjval->language_object;
			if(po) {
				Py_INCREF(po);
				return po;
			}
			break;
		}
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
pyro_convert_spec(Tk_ConfigSpec *spec, PyObject *args)
{
	/* see tkConfig.c:FormatConfigInfo */
	PyObject *rv;

	if(spec->type == TK_CONFIG_SYNONYM)
		rv = PyTuple_New(2);
	else
		rv = PyTuple_New(5);

	PyTuple_SetItem(rv, 0, PyString_FromString(spec->argvName));
	PyTuple_SetItem(rv, 1, PyString_FromString(
					spec->dbName ? spec->dbName: ""));
	if(spec->type == TK_CONFIG_SYNONYM)
		return rv;
	PyTuple_SetItem(rv, 2, PyString_FromString(
					spec->dbClass ? spec->dbClass : ""));
	PyTuple_SetItem(rv, 3, PyString_FromString(
					spec->defValue ? spec->defValue : ""));
	rivet_config_calc_result(spec);
	PyTuple_SetItem(rv, 4, pyro_conv_cget(0, args));

	return rv;

}

static PyObject *
pyro_conv_configure(char *str, PyObject *args)
{
	/* see tkConfig.c:Tk_ConfigureInfo */
	PyObject *rv;
	if(rivet_config_specs_len == 0) {
		Py_INCREF(Py_None);
		rv = Py_None;
	} else
	if(rivet_config_specs_len == 1) {
		/* get the config values for one spec. */
		rv = pyro_convert_spec(rivet_config_specs[0], args);
	} else {
		/* get the config values for multiple specs. */
		int i;
		PyObject *tmp;
		rv = PyTuple_New(rivet_config_specs_len);
		for(i=0;i<rivet_config_specs_len;i++) {
			tmp = pyro_convert_spec(rivet_config_specs[i], args);
			PyTuple_SetItem(rv, i, tmp);
		}
	}
		

	rivet_config_specs_len = 0; /* reset */
	return rv;
}

/*----------------------------------------------------------------------*/


static PyObject *
pyro_cmd(char *cmdname, char *method, int (*func)(), void *po,
	PyObject *args, PyObject *d, PyObject *(*convertfunc)())
{
	Argcv argcv;
	Rivetobj ro;
	int rv;
	ClientData cd;

	if(!rivet_main_window) {
		PyErr_SetString(pyro_error, not_inited);
		return NULL;
	}

	if(po && is_pyroobj((PyObject*)po)) {
		/*
		 * Acting as a method on a rivet object, assemble the
		 * arguments to look like "cmdname widgetname ?args?" ONLY.
		 * Method is always zero in this case, and the getattr
		 * function has already validated that the tk object exists.
		 */
		INIT_ARGCV(argcv, 2)
		argcv.argv[0] = cmdname;
		argcv.argv[1] = rivet_path(((Pyroobj)po)->rivetobj);
		ro = ((Pyroobj)po)->rivetobj;
		cd = ro->tkwin;
	} else {
		/*
		 * It's a module function, assemble the arguments
		 * to be "cmdname ?args?" or "cmdname method ?args?"
		 */
		INIT_ARGCV(argcv, (method ? 2 : 1))
		argcv.argv[0] = cmdname;
		if(method)
			argcv.argv[1] = method;
		ro = rivet_main_window;
		cd = po ? po : ro->tkwin;
	}

	if(pyro_parse_argcv(&argcv, args)) {
		return NULL;
	}
	if(d && pyro_parse_argcv(&argcv, d)) {
		return NULL;
	}

	Tcl_SetResult(ro->interp, 0, 0);
	rv = (func)(cd, ro->interp, argcv.argc, argcv.argv);
	FREE_ARGCV(argcv)
	if(rv == TCL_ERROR) {
		PyErr_SetString(pyro_error, ro->interp->result);
		return NULL;
	}
	if(ro->interp)
		return (convertfunc)(ro->interp->result, args);
	else {
		/* This happens when the main window is destroyed during a
		 * call to rivet.update and the default handler is installed.
		 */
		if(!rivet_main_window->interp) {
			XCloseDisplay(rivet_main_window->display);
			rivet_main_window = 0;
		}
		return (convertfunc)("", args);
	}
}

#define PYROCOMMAND( _cmdname, _pyrofunc, _tkfunc, _convert )\
	static PyObject *						\
	_pyrofunc (PyObject *po, PyObject *args, PyObject *d)		\
	{								\
	  return pyro_cmd( _cmdname, 0, _tkfunc, (Pyroobj) po,		\
				args, d, _convert);			\
	}

#define PYROARGCOMMAND( _cmdname, _tkmethod, _pyrofunc, _tkfunc, _convert )\
	static PyObject *						\
	_pyrofunc (PyObject *po, PyObject *args, PyObject *d)		\
	{								\
	  return pyro_cmd( _cmdname, _tkmethod, _tkfunc, 0,		\
				args, d, _convert);			\
	}


PYROCOMMAND( "lower",	pyro_lower,	Tk_LowerCmd, pyro_conv_none)
PYROCOMMAND( "raise",	pyro_raise,	Tk_RaiseCmd, pyro_conv_none)
PYROCOMMAND( "focus",	pyro_focus,	Tk_FocusCmd, pyro_conv_none)
PYROCOMMAND( "pack",	pyro_pack,	Tk_PackCmd, pyro_conv_str)
PYROCOMMAND( "place",	pyro_place,	Tk_PlaceCmd, pyro_conv_str)
PYROCOMMAND( "update",	pyro_update,	Tk_UpdateCmd, pyro_conv_none)
PYROARGCOMMAND("tk","appname",		\
			pyro_tk_appname, Tk_TkCmd, pyro_conv_str2)

/*----------------------------------------------------------------------*/
/* bindtags requires that all the tags, if spec'd, are concatenated
 * in one arg.  To allow passing in a list, we have to do the concat here.
 */
static int
pyro_bindtags_command(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
	char *tmp;
	char buffer[100];
	int i;
	if(argc>2) {
		buffer[0] = 0;
		for(i=2;i<argc;i++) {
			strncat(buffer, argv[i], sizeof(buffer)-1);
			strncat(buffer, " ", sizeof(buffer)-1);
		}
		argc = 3;
		argv[2] = buffer;
	}
	return Tk_BindtagsCmd(cd, interp, argc, argv);
}
	
static PyObject *
pyro_bindtags(PyObject *po, PyObject *args)
{
	if(PyTuple_Size(args) == 1)
		/* getting */
		return pyro_cmd("bindtags", 0, Tk_BindtagsCmd,
			(Pyroobj) po, args, 0, pyro_conv_tuple_rivetobj);
	else
		/* setting */
		return pyro_cmd("bindtags", 0, pyro_bindtags_command,
			(Pyroobj) po, args, 0, pyro_conv_none);
}

/*----------------------------------------------------------------------*/

PyObject *
pyro_new_callarg(Rivetobj rivetobj, XEvent *xevent, PyObject *arg, int keysym)
{
	int i, len;
	PyObject *callbackarg;
	PyObject *item, *expanditem;
	char *str;
	char *strval;
	int intval;

	len = PyTuple_Size(arg);
	callbackarg = PyTuple_New(len-1);
	if(!callbackarg)
		return 0; /* out of memory? */

	for(i=1; i<=len-1; i++) {
	    item = PyTuple_GetItem(arg, i);

	    if(PyString_Check(item)) {
		str = PyString_AsString(item);
		if(xevent && str[0] == '%' && str[1] && !str[2]) {
			if(str[1] == 'W') {
				item = rivetobj->language_object;
				Py_INCREF(item);
			} else {
				strval = rivet_expand_event(rivetobj,
						str+1, xevent, keysym, &intval);
				if(strval)
					item = PyString_FromString(strval);
				else
					item = PyInt_FromLong((long)intval);
			}
		} else {
			Py_INCREF(item);
		}
	    } else {
		Py_INCREF(item);
	    }
	    PyTuple_SetItem(callbackarg, i-1, item);
	}
	return callbackarg;
}

static int
pyro_call_callback(Rivetobj rivetobj, XEvent *xevent, ClientData assoc, int keysym, int callbacktype)
{
	PyObject *rv;
	PyObject *python_callback;
	static PyObject *s_tuple;
	PyObject *callbackobj;
	PyObject *callbackarg;

	if(!s_tuple)
		s_tuple = PyTuple_New(0);
	callbackobj = (PyObject*)assoc;

	/* The only things that are allowed to be in callbackobj
	 * are a callable, or a callable as the first item in a tuple.
	 */
	if(PyTuple_Check(callbackobj)) {
		python_callback = PyTuple_GetItem(callbackobj, 0);
		callbackarg = pyro_new_callarg(rivetobj, xevent,
				callbackobj, keysym);
		if(!callbackarg)
			return TCL_BREAK;
	} else {
		python_callback = callbackobj;
		callbackarg = s_tuple;
	}

	rv = PyEval_CallObject(python_callback, callbackarg);
	if(!rv)
		/* gack python error in execution, do something better here*/
		PyErr_Print();
	if(callbackarg != s_tuple)
		Py_DECREF(callbackarg);
	if(rv == Py_None) {
		Py_DECREF(rv);
		return TCL_OK;
	}
	if(rv)
		Py_DECREF(rv);
	return TCL_BREAK;
}

PyObject *
pyro_splice_additional_args(PyObject *arg1, PyObject *arg2, ClientData assoc)
{
	int i, l;
	PyObject *tuple;
	PyObject *tmp;

	/* arg1 must be set, arg2 may be 0 */

	if(PyTuple_Check((PyObject*)assoc)) {
		/* Copy the tuple to a new tuple, splicing in at index 1.
		 * Why?  Because this is a tuple, the first item is a
		 * callable, and it has to be first to be invoked by other
		 * routines.
		 */
		int offset;
		l = PyTuple_Size((PyObject*) assoc);
		tuple = PyTuple_New(l+ (arg2 ? 2 : 1));
		if(!tuple) {
			Py_DECREF(arg1);
			Py_DECREF(arg2);
			return 0;
		}
		tmp = PyTuple_GetItem((PyObject*)assoc, 0);
		PyTuple_SetItem(tuple, 0, tmp);
		Py_INCREF(tmp);
		PyTuple_SetItem(tuple, 1, arg1);
		if(arg2) {
			PyTuple_SetItem(tuple, 2, arg2);
			offset=2;
		} else {
			offset=1;
		}
		for(i=1; i<l; i++) {
			tmp = PyTuple_GetItem((PyObject*) assoc, i);
			PyTuple_SetItem(tuple, i+offset, tmp);
			Py_INCREF(tmp);
		}
	} else {
		/* This is the simple case where the assoc is a callable not
		 * wrapped in a tuple.  But to splice in the additional values,
		 * create a tuple. 
		 */
		if(!arg2)
			tuple = PyTuple_New(2);
		else
			tuple = PyTuple_New(3);

		if(!tuple) {
			Py_DECREF(arg1);
			Py_DECREF(arg2);
			return 0;
		}
		PyTuple_SetItem(tuple, 0, (PyObject*)assoc);
		PyTuple_SetItem(tuple, 1, arg1);
		if(arg2)
			PyTuple_SetItem(tuple, 2, arg2);
		Py_INCREF((PyObject*)assoc);
	}
	return tuple;
}

static PyObject *
pyro_bind_guts(Rivetobj rivetobj, char *classtag, char *pat, PyObject *callbackobj, int append)
{
	int rv;
	if(!pat) {
		/* must be get all bindings */
		char *str;
		str = rivet_get_all_bindings(rivetobj, classtag);
		if(str)
			return PyString_FromString(str);
		else
			Py_INCREF(Py_None);
			return Py_None;
	} else
	if(!callbackobj) {
		/* Just the event pattern was specified, do a get */
		Callback cb;
		int ncallbacks;
		int npy_callbacks;
		PyObject *tuple;
		int i, j;

		cb = rivet_get_binding(rivetobj, classtag, pat, &ncallbacks);

		for(i=0, npy_callbacks=0; i<ncallbacks; i++)
			if(cb[i].type == PYRO_UNIQUE_TOKEN)
				npy_callbacks++;
		tuple = PyTuple_New(npy_callbacks);
		if(!tuple) {
			return 0;
		}
		for(i=0, j=0; i<ncallbacks; i++)
		    if(cb[i].type == PYRO_UNIQUE_TOKEN) {
			PyTuple_SetItem(tuple, j, (PyObject*)cb[i].assoc);
			Py_INCREF((PyObject*)cb[i].assoc);
			j++;
		    }
		return tuple;
	} else
	if(callbackobj == Py_None) {
		/* Setting the binding to none, so delete it. */
		rv = rivet_delete_binding(rivetobj, classtag, pat);
		if(rv == TCL_ERROR) {
			PyErr_SetString(pyro_error,
				rivet_main_window->interp->result);
			return NULL;
		}
	} else {
		/* Add a binding. */
		if(!pyro_callable_check(callbackobj, 1))
			return NULL;

		/* 
		 * The tk bind stuff keeps track of reference counts,
		 * so we don't incref before passing the callbackobj along.
		 */
		rv = rivet_create_binding(rivetobj, classtag, pat,
				pyro_call_callback,
				(ClientData) callbackobj, append,
				PYRO_UNIQUE_TOKEN);
		if(rv == TCL_ERROR) {
			PyErr_SetString(pyro_error,
					rivet_main_window->interp->result);
			return NULL;
		}
	}

	Py_INCREF(Py_None);
	return Py_None;
}


static PyObject *
pyro_bind(PyObject *self, PyObject *args)
{
	char *class;
	char *pat;
	PyObject *callbackobj;
	int append;
	Pyroobj po;
	Rivetobj rivetobj;

	if(!rivet_main_window) {
		PyErr_SetString(pyro_error, not_inited);
		return NULL;
	}

	callbackobj = 0;
	class = 0;
	append = 0;
	pat = 0;
	if(self) {
		/* single object binding */
		po = (Pyroobj) self;
		if(PyArg_ParseTuple(args, "|sOi",
				&pat, &callbackobj, &append) == NULL)
			return NULL;
		rivetobj = po->rivetobj;

	} else {
		if(PyArg_ParseTuple(args, "s|sOi", &class,
				&pat, &callbackobj, &append) != NULL) {
			/* class binding */
			rivetobj = 0;
		} else
		if(PyArg_ParseTuple(args, "O|sOi",
				&po, &pat, &callbackobj, &append) != NULL) {
			/* single object binding */
			po = pyro_get_widget(po);
			if(!po)
				return NULL;
			rivetobj = po->rivetobj;
		} else
			return NULL;
	}
	return pyro_bind_guts(rivetobj, class, pat, callbackobj, append);
}

static PyObject *
pyro_tag_bind(PyObject *self, PyObject *args)
{
	char *tag;
	char *pat;
	PyObject *callbackobj;
	int append;
	Pyroobj po;
	Rivetobj rivetobj;

	po = (Pyroobj) self;

	callbackobj = 0;
	append = 0;
	pat = 0;
	if(PyArg_ParseTuple(args, "s|sOi",
				&tag, &pat, &callbackobj, &append) == NULL)
			return NULL;

	return pyro_bind_guts(po->rivetobj, tag, pat, callbackobj, append);
}

static int
pyro_call_callback2(Rivetobj rivetobj, XEvent *xevent, ClientData assoc, ClientData misc, int callbacktype)
{
	/* This function handles the variations in widget callbacks.
	 * Especially the different uses of the 4th parameter.
	 */
	PyObject *tuple;
	int rv;

	switch (callbacktype) {
	  case RIVET_CB_BUTTON_INVOKE: {
		/* unlike tk, allow getting to event info from a
		 * button command.
		 */
		extern XEvent *rivet_button_invoke_event;
		xevent = rivet_button_invoke_event;
		rv = pyro_call_callback(rivetobj, xevent, assoc,
			0, callbacktype);
		break;
		}

	  case RIVET_CB_SCALE_INVOKE: {
		/* 1) No xevent is available for scales, a exception will
		 *    be generated if the callback tries to get event
		 *    details besides the object '%W'.
		 * 2) a float object is created to pass the data on,
		 *    we have to make this similar to other callbacks
		 *    and pack it in a tuple.
		 */
		PyObject *dbl;

		dbl = PyFloat_FromDouble(*(double*)misc);
		if(!dbl)
			return TCL_ERROR;

		tuple = pyro_splice_additional_args(dbl, 0, assoc);
		if(!tuple)
			return TCL_ERROR;
		rv = pyro_call_callback(rivetobj, 0, (ClientData) tuple,
			0, callbacktype);
		Py_DECREF(tuple);
		break;
		}
	  case RIVET_CB_SBAR_CLIENT_CHANGED: {
		PyObject *dbl1;
		PyObject *dbl2;
		double *d = (double*)misc;

		dbl1 = PyFloat_FromDouble(d[0]);
		dbl2 = PyFloat_FromDouble(d[1]);
		if(!dbl1 || !dbl2)
			return TCL_ERROR;

		/* Splice in the first,last arguments. */
		tuple = pyro_splice_additional_args(dbl1, dbl2, assoc);
		if(!tuple)
			return TCL_ERROR;
		rv = pyro_call_callback(rivetobj, 0, (ClientData) tuple,
			0, callbacktype);
		Py_DECREF(tuple);
		break;
		}
	  case RIVET_CB_SBAR_CHANGE_CLIENT_YVIEW:
	  case RIVET_CB_SBAR_CHANGE_CLIENT_XVIEW: {
		PyObject *scrollargs;
		Scrollbar_notify_data *snd = (Scrollbar_notify_data*)misc;

		switch(snd->scroll_op) {
		  case TK_SCROLL_MOVETO:
			scrollargs = Py_BuildValue("(sd)",
				"moveto", snd->newpos);
			break;
		  case TK_SCROLL_PAGES:
			scrollargs = Py_BuildValue("(sds)",
				"scroll", snd->newpos, "pages");
			break;
		  case TK_SCROLL_UNITS:
			scrollargs = Py_BuildValue("(sds)",
				"scroll", snd->newpos, "units");
			break;
		}

		if(!scrollargs)
			return TCL_ERROR;

		/* Splice in the first,last arguments. */
		tuple = pyro_splice_additional_args(scrollargs, 0, assoc);
		if(!tuple)
			return TCL_ERROR;
		rv = pyro_call_callback(rivetobj, 0, (ClientData) tuple,
				0, callbacktype);
		Py_DECREF(tuple);
		break;
		}
	  default:
		rv = pyro_call_callback(rivetobj, xevent, assoc,
			(int)misc, callbacktype);
	}
	return rv;
}

/*----------------------------------------------------------------------*/

typedef struct {
	PyObject_HEAD
	Tk_TimerToken token;
	ClientData assoc;
} Pyro_timer_struct;

staticforward PyTypeObject PyroTimer;

static void
pyro_after_call_callback(ClientData clientdata)
{
	pyro_call_callback(0, 0, clientdata, 0, RIVET_CB_TIMER);
	Py_DECREF((PyObject*)clientdata);
}

static PyObject *
pyro_after(PyObject *self, PyObject *args)
{
	Pyroobj po;
	int delay;
	char *idlestr;
	PyObject *callbackobj;
	Pyro_timer_struct *tt;

	if(!rivet_main_window) {
		PyErr_SetString(pyro_error, not_inited);
		return NULL;
	}

	callbackobj = 0;
	idlestr = 0;
	if(PyArg_ParseTuple(args, "i|O", &delay, &callbackobj) == NULL)
	  if(PyArg_ParseTuple(args, "sO", &idlestr, &callbackobj) == NULL)
		return NULL;

	PyErr_Clear();

	if(!callbackobj && !idlestr) {
		Tk_Sleep(delay);
		Py_INCREF(Py_None);
		return Py_None;
	}

	if(!pyro_callable_check(callbackobj, 1))
		return NULL;

	tt = PyObject_NEW(Pyro_timer_struct, &PyroTimer);
	if(!tt)
		return NULL;

	Py_INCREF(callbackobj);

	if(idlestr && !strcmp(idlestr, "idle")) {
		Tk_DoWhenIdle(pyro_after_call_callback, callbackobj);
		tt->token = (Tk_TimerToken)-1; /* -1 flags cancel idle proc */
		tt->assoc = callbackobj;
	} else {
		tt->token = Tk_CreateTimerHandler(delay,
					pyro_after_call_callback, callbackobj);
		tt->assoc = 0;
	}
	return (PyObject*)tt;
}

static PyObject *
pyro_after_cancel(PyObject *self, PyObject *args)
{
	Pyro_timer_struct *tt;
	if(self->ob_type != &PyroTimer) {
		PyErr_SetString(pyro_error, "Not a timer object");
		return NULL;
	}
	tt = (Pyro_timer_struct*) self;
	if(tt->token == (Tk_TimerToken)-1)
		Tk_CancelIdleCall(pyro_after_call_callback, tt->assoc);
	else
		Tk_DeleteTimerHandler(tt->token);
	Py_INCREF(Py_None);
	return(Py_None);
}

static PyObject *
pyrotimer_getattr(Pyroobj po, char *name)
{
	static PyMethodDef timer_methods[] = {
		{"cancel", pyro_after_cancel, 1,
				"Stop a timer before expiration"},
		{0},
	};
	return Py_FindMethod(timer_methods, (PyObject*)po, name);
}

static void dummy_dealloc(PyObject *p) {}

static PyTypeObject PyroTimer = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,				/* size */
	"timer_token",			/* name */
	sizeof(Pyro_timer_struct),	/* size */
	0,				/* itemsize */
	(destructor)dummy_dealloc,	/* dealloc */
	0,				/* print */
	(getattrfunc)pyrotimer_getattr,	/* getattr */
	0,				/* setattr */
	0,				/* compare */
	0,				/* repr */
	0,				/* as number */
	0,				/* as sequence */
	0, 				/* as mapping */
};

/*----------------------------------------------------------------------*/
/* begin variables */

static PyObject *
pyro_getvar(PyObject *ignored, PyObject *args)
{
	char *name;
	char *val;

	if(!rivet_main_window) {
		PyErr_SetString(pyro_error, not_inited);
		return NULL;
	}

	if(PyArg_ParseTuple(args, "s", &name) == NULL)
		return NULL;

	Tcl_SetResult(rivet_main_window->interp, 0, 0);
	val = Tcl_GetVar(rivet_main_window->interp,
			name, TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG);
	if(!val) {
		PyErr_SetString(pyro_error, rivet_main_window->interp->result);
		return NULL;
	}
	return PyString_FromString(val);
}

static PyObject *
pyro_setvar(PyObject *ignored, PyObject *args)
{
	char *name;
	char *val;

	if(!rivet_main_window) {
		PyErr_SetString(pyro_error, not_inited);
		return NULL;
	}

	if(PyArg_ParseTuple(args, "ss", &name, &val) == NULL)
		return NULL;

	Tcl_SetResult(rivet_main_window->interp, 0, 0);
	val = Tcl_SetVar(rivet_main_window->interp,
			name, val, TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG);
	if(!val) {
		PyErr_SetString(pyro_error, rivet_main_window->interp->result);
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
pyro_unsetvar(PyObject *ignored, PyObject *args)
{
	char *name;
	int rv;

	if(!rivet_main_window) {
		PyErr_SetString(pyro_error, not_inited);
		return NULL;
	}

	if(PyArg_ParseTuple(args, "s", &name) == NULL)
		return NULL;

	Tcl_SetResult(rivet_main_window->interp, 0, 0);
	rv = Tcl_UnsetVar(rivet_main_window->interp,
			name, TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG);
	if(rv == TCL_ERROR) {
		PyErr_SetString(pyro_error, rivet_main_window->interp->result);
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static char *
pyro_trace_callback(ClientData clientdata, Tcl_Interp *interp,
		char *name, char *unused, int flags)
{
	PyObject *str;
	int result;

	str = PyString_FromString(name);
	if(!str)
		return NULL;

	result = pyro_call_callback(0, 0, clientdata, 0, RIVET_CB_TRACE);

	if(flags & TCL_TRACE_DESTROYED) {
		PyObject *obj = (PyObject *)clientdata;
		Py_DECREF(obj);
	}
	if(result != TCL_OK)
		return "Stop trace.";
	return NULL; /* no error */
}

static PyObject *
pyro_tracevar(PyObject *ignored, PyObject *args)
{
	char *name;
	char *flags;
	int flagsi;
	PyObject *callbackobj;
	int rv;

	if(!rivet_main_window) {
		PyErr_SetString(pyro_error, not_inited);
		return NULL;
	}

	if(PyArg_ParseTuple(args, "ssO", &name, &flags, &callbackobj) == NULL)
		return NULL;

	if(!pyro_callable_check(callbackobj, 1))
		return NULL;

	flagsi = (TCL_GLOBAL_ONLY|TCL_TRACE_UNSETS);
	for(; *flags; flags++)
		switch(*flags) {
		case 'r':
		case 'R':
			flagsi |= TCL_TRACE_READS;
			break;
		case 'w':
		case 'W':
			flagsi |= TCL_TRACE_WRITES;
			break;
		}
	if(flagsi == (TCL_GLOBAL_ONLY|TCL_TRACE_UNSETS)) {
		PyErr_SetString(pyro_error, "no flags specified");
		return NULL;
	}

	Py_INCREF(callbackobj);

	Tcl_SetResult(rivet_main_window->interp, 0, 0);
	rv = Tcl_TraceVar(rivet_main_window->interp,
			name, flagsi, pyro_trace_callback, callbackobj);
	if(rv == TCL_ERROR) {
		Py_DECREF(callbackobj);
		PyErr_SetString(pyro_error, rivet_main_window->interp->result);
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}


static PyObject *
pyro_untracevar(PyObject *ignored, PyObject *args)
{
	/* warning: this is a low level function, if the callback 
	 * to be untraced is not previously traced, then the callback
	 * will be unnecessarily decref'd.  Suggestion: use the rivetutil.py
	 * variable types which use this interface correctly.
	 */
	char *name;
	char *flags;
	int flagsi;
	PyObject *callbackobj;
	PyObject *tmp;

	if(!rivet_main_window) {
		PyErr_SetString(pyro_error, not_inited);
		return NULL;
	}

	if(PyArg_ParseTuple(args, "ssO", &name, &flags, &callbackobj) == NULL)
		return NULL;

	if(!pyro_callable_check(callbackobj, 1))
		return NULL;

	flagsi = (TCL_GLOBAL_ONLY|TCL_TRACE_UNSETS);
	for(; *flags; flags++)
		switch(*flags) {
			break;
		case 'r':
		case 'R':
			flagsi |= TCL_TRACE_READS;
			break;
		case 'w':
		case 'W':
			flagsi |= TCL_TRACE_WRITES;
			break;
		}
	if(flagsi == (TCL_GLOBAL_ONLY|TCL_TRACE_UNSETS)) {
		PyErr_SetString(pyro_error, "specify R or W for flags");
		return NULL;
	}


	/* The following loop makes sure that there is indeed a trace
	 * with the right specifications.  This is important, otherwise
	 * we'll decref the object when we shouldn't.
	 */
	tmp = NULL;
	do {
		tmp = (PyObject*) Tcl_VarTraceInfo(rivet_main_window->interp,
			name, flagsi, pyro_trace_callback, (ClientData)tmp);
		if(tmp == callbackobj)
			/* found it, proceed to remove it */
			break;
		if(tmp &&
		   PyMethod_Check(tmp) &&
		   PyMethod_Check(callbackobj) &&
		   PyMethod_Self(tmp) == PyMethod_Self(callbackobj) &&
		   PyMethod_Function(tmp) == PyMethod_Function(callbackobj))
			/* found it, proceed to remove it */
			/* Instance-method objects are a common case, so
			 * check for it. */
			break;
	} while(tmp);
	if(!tmp) {
		PyErr_SetString(pyro_error,
		"variable is not traced by that function and/or with that mode");
		return NULL;
	}

	Tcl_SetResult(rivet_main_window->interp, 0, 0);
	Tcl_UntraceVar(rivet_main_window->interp,
			name, flagsi, pyro_trace_callback, callbackobj);
	Py_DECREF(callbackobj); /* Must be correct object because untracevar
				 * returns no status.  */
	Py_INCREF(Py_None);
	return Py_None;
}

/* end variables */
/*----------------------------------------------------------------------*/


static PyObject *
pyro_bell(PyObject *ignored, PyObject *args)
{
	Pyroobj po;

	if(!rivet_main_window) {
		PyErr_SetString(pyro_error, not_inited);
		return NULL;
	}

	if(PyArg_ParseTuple(args, "O", &po) != NULL) {
		po = pyro_get_widget(po);
		if(!po)
			return NULL;
	} else {
		po = (Pyroobj)rivet_main_window->language_object;
	}
	PyErr_Clear();

	XBell(Tk_Display(po->rivetobj->tkwin), 0);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
pyro_main_widget(PyObject *ignored, PyObject *args)
{
	Pyroobj po;

	if(!rivet_main_window) {
		PyErr_SetString(pyro_error, not_inited);
		return NULL;
	}

	po = (Pyroobj) rivet_main_window->language_object;
	Py_INCREF((PyObject*)po);
	return (PyObject*)po;
}

static PyObject *
pyro_parent(Pyroobj po, PyObject *args)
{
	Rivetobj parent;
	Pyroobj pparent;

	if(!is_pyroobj(po))
		return NULL;

	parent = rivet_parent(po->rivetobj);
	pparent = (Pyroobj) (parent->language_object);
	Py_INCREF(pparent);
	return (PyObject*) pparent;
}

static PyObject *
pyro_children(Pyroobj po, PyObject *args)
{
	PyObject *tuple;
	PyObject *item;
	TkWindow *win;
	int len;
	int i;

	if(!is_pyroobj(po))
		return NULL;

	len = 0;
	for(win=po->rivetobj->tkwin->childList; win ; win = win->nextPtr)
		len++;

	if(!len) {
		Py_INCREF(Py_None);
		return Py_None;
	}

	tuple = PyTuple_New(len);
	if(!tuple) {
		Py_INCREF(Py_None);
		return Py_None;
	}
	i=0;
	for(win=po->rivetobj->tkwin->childList; win ; win = win->nextPtr) {
		item = (PyObject*)win->rivetobj->language_object;
		PyTuple_SetItem(tuple, i, item);
		Py_INCREF(item);
		i++;
	}
	return tuple;
}

static PyObject *
pyro_name(Pyroobj self, PyObject *args)
{
	Pyroobj po;
	
	po = pyro_get_widget(self);
	if(!po)
		return(NULL);
	return PyString_FromString(rivet_path(po->rivetobj));
}


static PyObject *
pyro_fileevent(PyObject *po, PyObject *args)
{
	PyObject *rv;
	int fileno;
	int len;
	char *mode;
	int imode;
	Callback cb;
	PyObject *callbackobj;

	if(PyArg_ParseTuple(args, "is|O", &fileno, &mode, &callbackobj) == NULL)
		return NULL;

	len = strlen(mode);
	if(strncmp("readable", mode, len)==0)
		imode = 0;
	else
	if(strncmp("writeable", mode, len)==0)
		imode = 1;

	if(PyTuple_Size(args) == 2) {
		/* get the handler for this fileno, imode */
		cb = rivet_get_fileevent(fileno, imode);
		if(!cb || cb->type != PYRO_UNIQUE_TOKEN) {
			if(cb)
				rivet_decref_callback(cb);
			Py_INCREF(Py_None);
			return Py_None;
		}
		return cb->assoc; /* already incref'd by rivet_get_fileevent */
	}

	/* assume (PyTuple_Size(args) == 3) */
	if(callbackobj == Py_None)
		rivet_set_fileevent(fileno, imode, (Callback)"");
	else
		rivet_set_fileevent(fileno, imode,
			(Callback)rivet_new_callback(pyro_call_callback2,
				callbackobj, PYRO_UNIQUE_TOKEN));

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
pyro_mainloop(PyObject *self, PyObject *args)
{
	if(!rivet_main_window || !rivet_main_window->interp) {
		PyErr_SetString(pyro_error, not_inited);
		return NULL;
	}
		
	Tcl_SetResult(rivet_main_window->interp, 0, 0);
	Tk_MainLoop();

	if(rivet_main_window && rivet_main_window->display) {
		XCloseDisplay(rivet_main_window->display);
		rivet_main_window = 0;
	}

	Py_INCREF(Py_None);
	return Py_None;
}


/*----------------------------------------------------------------------*/
/* begin images */


/* This macro is similar to PYROCOMMAND, but does different things with
 * the args so that pyro_cmd will work with image objects.
 */
#define PYROIMAGECOMMAND(_tkmethod, _pyrofunc, _convert )\
	static PyObject *				\
	_pyrofunc (PyObject *po, PyObject *args, PyObject *d)\
	{						\
	  char *arg0; /* i.e. the image name. */	\
	  Pyroimage pi = (Pyroimage) po;		\
	  if(!pi->name)	{				\
		PyErr_SetString(pyro_error, "Image has been deleted");\
		return NULL;				\
	  }						\
	  return pyro_cmd(PyString_AsString(pi->name), _tkmethod, \
			 pi->func, (Pyroobj) pi->data, args, d, _convert);\
        }

PYROIMAGECOMMAND("blank",	pyro_image_blank, pyro_conv_str)
PYROIMAGECOMMAND("cget",	pyro_image_cget, pyro_conv_cget)
PYROIMAGECOMMAND("configure",	pyro_image_configure, pyro_conv_configure)
PYROIMAGECOMMAND("copy",	pyro_image_copy, pyro_conv_str)
PYROIMAGECOMMAND("get",		pyro_image_get, pyro_conv_tuple_int)
PYROIMAGECOMMAND("put",		pyro_image_put, pyro_conv_str)
PYROIMAGECOMMAND("read",	pyro_image_read, pyro_conv_str)
PYROIMAGECOMMAND("redither",	pyro_image_redither, pyro_conv_none)
PYROIMAGECOMMAND("write",	pyro_image_write, pyro_conv_str)

static int pyro_image_command(ClientData, Tcl_Interp *, int , char **);

PYROARGCOMMAND( "image","width",	\
			pyro_image_width, Tk_ImageCmd, pyro_conv_int)
PYROARGCOMMAND( "image","height",	\
			pyro_image_height, Tk_ImageCmd, pyro_conv_int)
PYROARGCOMMAND( "image","type",	\
			pyro_image_type, Tk_ImageCmd, pyro_conv_str)
PYROARGCOMMAND( "image","types",	\
			pyro_image_types, Tk_ImageCmd, pyro_conv_tuple_str)
PYROARGCOMMAND( "image","create",	\
			pyro_image_create, pyro_image_command, \
			pyro_conv_passthrough)
PYROARGCOMMAND( "image","delete",	\
			pyro_image_delete, pyro_image_command, pyro_conv_none)


static PyObject *imagedict;

static void
pyroimage_dealloc(Pyroimage pi)
{
	char *argv[3];

	if(!pi || !pi->name)
		return; /* tk object already deleted */

	/* clean up tk's image object */
	argv[0] = "image";
	argv[1] = "delete";
	argv[2] = PyString_AsString(pi->name);
	Tcl_SetResult(rivet_main_window->interp, 0, 0);
	Tk_ImageCmd(rivet_main_window->tkwin, rivet_main_window->interp,
			3, argv);

	/* clean up name reference */
	Py_DECREF(pi->name); /* for pi->name */
}

static PyObject *
pyro_image_names(PyObject *obj, PyObject *args)
{
	if(!imagedict)
		imagedict = PyDict_New();
	Py_INCREF(imagedict);
	return imagedict;
}

static int
pyro_image_command(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
	/* Wrapper to tk's image function, 
	 * only used for creating and deleting images.
	 */
	int rv;
	Pyroimage pi;

	static PyMethodDef photo_methods[] =  {
		{"blank",	(PyCFunction)pyro_image_blank,
						METH_VARARGS|METH_KEYWORDS},
		{"cget",	(PyCFunction)pyro_image_cget,
						METH_VARARGS|METH_KEYWORDS},
		{"configure",	(PyCFunction)pyro_image_configure,
						METH_VARARGS|METH_KEYWORDS},
		{"copy",	(PyCFunction)pyro_image_copy,
						METH_VARARGS|METH_KEYWORDS},
		{"get",		(PyCFunction)pyro_image_get,
						METH_VARARGS|METH_KEYWORDS},
		{"put",		(PyCFunction)pyro_image_put,
						METH_VARARGS|METH_KEYWORDS},
		{"read",	(PyCFunction)pyro_image_read,
						METH_VARARGS|METH_KEYWORDS},
		{"redither",	(PyCFunction)pyro_image_redither,
						METH_VARARGS|METH_KEYWORDS},
		{"write",	(PyCFunction)pyro_image_write,
						METH_VARARGS|METH_KEYWORDS},
		{0}
	};
	static PyMethodDef bitmap_methods[] =  {
		{"cget",	(PyCFunction)pyro_image_cget,
						METH_VARARGS|METH_KEYWORDS},
		{"configure",	(PyCFunction)pyro_image_configure,
						METH_VARARGS|METH_KEYWORDS},
		{0}
	};

	rv = Tk_ImageCmd(cd, interp, argc, argv);
	if(rv == TCL_ERROR)
		return rv;

	if(!imagedict)
		imagedict = PyDict_New();

	if(!strcmp(argv[1], "create")) {
		/* Must be careful not to delete an image that is already
		 * in the dictionary. */
		pi = (Pyroimage)PyDict_GetItemString(imagedict, interp->result);
		
		if(!pi) {
			PyErr_Clear();
			pi = PyObject_NEW(Pyro_image_struct, &PyroImageType);
			if(!pi) 
				return TCL_ERROR;
			pi->name = PyString_FromString(interp->result);
			PyDict_SetItem(imagedict, pi->name, (PyObject*)pi);
		} else
			Py_INCREF(pi); /* for the return value */

		pi->func = rivetcmd_proc;
		pi->data = rivetcmd_clientdata;
		pi->methods = strcmp(argv[2], "photo") ?
					bitmap_methods : photo_methods;

		interp->result = (char*) pi; /* use native conversion */
	} else {
		/* must be doing a delete */
		int i;
		PyObject *name;
		
		for(i=2;i<argc;i++) {
		  Pyroimage pi2;

		  pi = (Pyroimage)PyDict_GetItemString(imagedict, argv[i]);
		  if(!pi || !pi->name)
			continue; /* perhaps already deleted */
		  name = pi->name;
		  pi->name = 0;
		  PyDict_DelItem(imagedict, name);
		  Py_DECREF(name); /* for the pi reference to name. */
		}
	}
	return rv;
}

static PyObject *
pyro_image_getattr(PyObject *obj, char *name)
{
	Pyroimage pi = (Pyroimage) obj;
	return Py_FindMethod(pi->methods, obj, name);
}

static PyTypeObject PyroImageType = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,				/* size */
	"rivet image",			/* name */
	sizeof(Pyro_image_struct),	/* size */
        0,				/* itemsize */
	(destructor)pyroimage_dealloc, 	/* dealloc */
        0,				/* print */
        (getattrfunc)pyro_image_getattr,/* getattr */
        (setattrfunc)0,			/* setattr */
        0,				/* compare */
        0,				/* repr */
        0,				/* as number */
        0,				/* as sequence */
        0,				/* as mapping */
        0,				/* tp_hash */
        (ternaryfunc)0,			/* tp_call */
        0,				/* tp_str */
};

static void
pyro_init_image_ops(PyObject *dict)
{
	Pyro_ms ms;
	static Pyrometh methods[] =  {
		PYROMDECL("create", pyro_image_create),
		PYROMDECL("delete", pyro_image_delete),
		PYROMDECL("height", pyro_image_height),
		PYROMDECL("width", pyro_image_width),
		PYROMDECL("names", pyro_image_names),
		PYROMDECL("types", pyro_image_types),
		PYROMDECL("type", pyro_image_type),
		{0}
	};

	ms = PyObject_NEW(Pyro_method_set_struct, &PyroMsType);
	ms->methods = pyro_init_methods(methods);
	ms->callfn = 0;

	PyDict_SetItemString(dict, "image", (PyObject *)ms);
}


/* end images */
/*----------------------------------------------------------------------*/
/* begin clipboard */

PYROARGCOMMAND("clipboard","clear",     \
                        pyro_clipboard_clear, Tk_ClipboardCmd, pyro_conv_none)
PYROARGCOMMAND("clipboard","append",    \
                        pyro_clipboard_append, Tk_ClipboardCmd, pyro_conv_none)
static void
pyro_init_clipboard_ops(PyObject *dict)
{
	Pyro_ms ms;
	static Pyrometh methods[] =  {
		PYROMDECL("clear", pyro_clipboard_clear),
		PYROMDECL("append", pyro_clipboard_append),
		{0}
	};

	ms = PyObject_NEW(Pyro_method_set_struct, &PyroMsType);
	ms->methods = pyro_init_methods(methods);
	ms->callfn = 0;

	PyDict_SetItemString(dict, "clipboard", (PyObject *)ms);
}
/* end clipboard */
/*----------------------------------------------------------------------*/
/* begin grab */

PYROARGCOMMAND("grab","current",     \
                        pyro_grab_current, Tk_GrabCmd, pyro_conv_tuple_rivetobj)
PYROARGCOMMAND("grab","release",     \
                        pyro_grab_release, Tk_GrabCmd, pyro_conv_none)
PYROARGCOMMAND("grab","set",     \
                        pyro_grab_set, Tk_GrabCmd, pyro_conv_none)
PYROARGCOMMAND("grab","status",     \
                        pyro_grab_status, Tk_GrabCmd, pyro_conv_str)
static void
pyro_init_grab_ops(PyObject *dict)
{
	Pyro_ms ms;
	static Pyrometh methods[] =  {
		PYROMDECL("current", pyro_grab_current),
		PYROMDECL("release", pyro_grab_release),
		PYROMDECL("set", pyro_grab_set),
		PYROMDECL("status", pyro_grab_status),
		{0}
	};

	ms = PyObject_NEW(Pyro_method_set_struct, &PyroMsType);
	ms->methods = pyro_init_methods(methods);
	ms->callfn = 0;

	PyDict_SetItemString(dict, "grab", (PyObject *)ms);
}
/* end grab */
/*----------------------------------------------------------------------*/
/* begin option */

PYROARGCOMMAND("option","add",     \
                        pyro_option_add, Tk_OptionCmd, pyro_conv_none)
PYROARGCOMMAND("option","clear",     \
                        pyro_option_clear, Tk_OptionCmd, pyro_conv_none)
PYROARGCOMMAND("option","get",     \
                        pyro_option_get, Tk_OptionCmd, pyro_conv_str)
PYROARGCOMMAND("option","readfile",     \
                        pyro_option_readfile, Tk_OptionCmd, pyro_conv_none)
static void
pyro_init_option_ops(PyObject *dict)
{
	Pyro_ms ms;
	static Pyrometh methods[] =  {
		PYROMDECL("add", pyro_option_add),
		PYROMDECL("clear", pyro_option_clear),
		PYROMDECL("get", pyro_option_get),
		PYROMDECL("readfile", pyro_option_readfile),
		{0}
	};

	ms = PyObject_NEW(Pyro_method_set_struct, &PyroMsType);
	ms->methods = pyro_init_methods(methods);
	ms->callfn = 0;

	PyDict_SetItemString(dict, "option", (PyObject *)ms);
}
/* end option */
/*----------------------------------------------------------------------*/
/* begin pack */

PYROARGCOMMAND("pack",0,	\
			pyro_pack_call,	Tk_PackCmd, pyro_conv_none)
PYROARGCOMMAND("pack","forget",     \
                        pyro_pack_forget, Tk_PackCmd, pyro_conv_none)
PYROARGCOMMAND("pack","info",     \
                        pyro_pack_info, Tk_PackCmd, pyro_conv_tuple_rivetobj)
PYROARGCOMMAND("pack","propagate",     \
                        pyro_pack_propagate, Tk_PackCmd, pyro_conv_str)
PYROARGCOMMAND("pack","slaves",     \
                        pyro_pack_slaves, Tk_PackCmd, pyro_conv_tuple_rivetobj)

static void
pyro_init_pack_ops(PyObject *dict)
{
	Pyro_ms ms;
	static Pyrometh methods[] =  {
		PYROMDECL("forget", pyro_pack_forget),
		PYROMDECL("info", pyro_pack_info),
		PYROMDECL("propagate", pyro_pack_propagate),
		PYROMDECL("slaves", pyro_pack_slaves),
		{0}
	};

	ms = PyObject_NEW(Pyro_method_set_struct, &PyroMsType);
	ms->methods = pyro_init_methods(methods);
	ms->callfn = pyro_pack_call;

	PyDict_SetItemString(dict, "pack", (PyObject *)ms);
}
/* end pack */
/*----------------------------------------------------------------------*/
/* begin place */

PYROARGCOMMAND("place",0,	\
			pyro_place_call, Tk_PlaceCmd, pyro_conv_none)
PYROARGCOMMAND("place","forget",	\
			pyro_place_forget, Tk_PlaceCmd, pyro_conv_none)
PYROARGCOMMAND("place","info",	\
			pyro_place_info, Tk_PlaceCmd, pyro_conv_tuple_rivetobj)
PYROARGCOMMAND("place","propagate",	\
			pyro_place_propagate, Tk_PlaceCmd, pyro_conv_str)
PYROARGCOMMAND("place","slaves",	\
			pyro_place_slaves, Tk_PlaceCmd, pyro_conv_tuple_rivetobj)
static void
pyro_init_place_ops(PyObject *dict)
{
	Pyro_ms ms;
	static Pyrometh methods[] =  {
		PYROMDECL("forget", pyro_place_forget),
		PYROMDECL("info", pyro_place_info),
		PYROMDECL("propagate", pyro_place_propagate),
		PYROMDECL("slaves", pyro_place_slaves),
		{0}
	};

	ms = PyObject_NEW(Pyro_method_set_struct, &PyroMsType);
	ms->methods = pyro_init_methods(methods);
	ms->callfn = pyro_place_call;

	PyDict_SetItemString(dict, "place", (PyObject *)ms);
}
/* end place */
/*----------------------------------------------------------------------*/
/* begin selection */

PYROARGCOMMAND("selection","clear",     \
                        pyro_selection_clear, Tk_SelectionCmd, pyro_conv_none)
PYROARGCOMMAND("selection","get",     \
                        pyro_selection_get, Tk_SelectionCmd, pyro_conv_str2)
PYROARGCOMMAND("selection","handle",     \
                        pyro_selection_handle, Tk_SelectionCmd, pyro_conv_none)
PYROARGCOMMAND("selection","own",     \
                        pyro_selection_own, Tk_SelectionCmd, pyro_conv_none)

static void
pyro_init_selection_ops(PyObject *dict)
{
	Pyro_ms ms;
	static Pyrometh methods[] =  {
		PYROMDECL("clear", pyro_selection_clear),
		PYROMDECL("get", pyro_selection_get),
		PYROMDECL("handle", pyro_selection_handle),
		PYROMDECL("own", pyro_selection_own),
		{0}
	};

	ms = PyObject_NEW(Pyro_method_set_struct, &PyroMsType);
	ms->methods = pyro_init_methods(methods);
	ms->callfn = 0;

	PyDict_SetItemString(dict, "selection", (PyObject *)ms);
}
/* end seletion */
/*----------------------------------------------------------------------*/
/* begin tkwait */

PYROARGCOMMAND("tkwait","variable",     \
                        pyro_tkwait_variable, Tk_TkwaitCmd, pyro_conv_none)
PYROARGCOMMAND("tkwait","visibility",     \
                        pyro_tkwait_visibility, Tk_TkwaitCmd, pyro_conv_none)
PYROARGCOMMAND("tkwait","window",     \
                        pyro_tkwait_window, Tk_TkwaitCmd, pyro_conv_none)

static void
pyro_init_tkwait_ops(PyObject *dict)
{
	Pyro_ms ms;
	static Pyrometh methods[] =  {
		PYROMDECL("variable", pyro_tkwait_variable),
		PYROMDECL("visibility", pyro_tkwait_visibility),
		PYROMDECL("window", pyro_tkwait_window),
		{0}
	};

	ms = PyObject_NEW(Pyro_method_set_struct, &PyroMsType);
	ms->methods = pyro_init_methods(methods);
	ms->callfn = 0;

	PyDict_SetItemString(dict, "tkwait", (PyObject *)ms);
}
/* end tkwait */
/*----------------------------------------------------------------------*/
/* begin winfo */

PYROARGCOMMAND("winfo","atom",			\
			pyro_winfo_atom, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","atomname",		\
			pyro_winfo_atomname, Tk_WinfoCmd, pyro_conv_str2)
PYROARGCOMMAND("winfo","cells",			\
			pyro_winfo_cells, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","children",		\
			pyro_winfo_children, Tk_WinfoCmd, pyro_conv_tuple_rivetobj)
PYROARGCOMMAND("winfo","class",			\
			pyro_winfo_class, Tk_WinfoCmd, pyro_conv_str2)
PYROARGCOMMAND("winfo","containing",		\
			pyro_winfo_containing, Tk_WinfoCmd, pyro_conv_rivetobj)
PYROARGCOMMAND("winfo","colormapfull",		\
			pyro_winfo_colormapfull, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","depth",			\
			pyro_winfo_depth, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","exists",		\
			pyro_winfo_exists, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","fpixels",		\
			pyro_winfo_fpixels, Tk_WinfoCmd, pyro_conv_float)
PYROARGCOMMAND("winfo","geometry",		\
			pyro_winfo_geometry, Tk_WinfoCmd, pyro_conv_str2)
PYROARGCOMMAND("winfo","height",		\
			pyro_winfo_height, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","id",		\
			pyro_winfo_id, Tk_WinfoCmd, pyro_conv_str2)
/* unimplemented
PYROARGCOMMAND("winfo","interps",		\
			pyro_winfo_interps, Tk_WinfoCmd, pyro_conv_none)
*/
PYROARGCOMMAND("winfo","ismapped",		\
			pyro_winfo_ismapped, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","manager",		\
			pyro_winfo_manager, Tk_WinfoCmd, pyro_conv_str)
/* unimplemented...
PYROARGCOMMAND("winfo","name",			\
			pyro_winfo_name, Tk_WinfoCmd, pyro_conv_str2)
*/
PYROARGCOMMAND("winfo","parent",		\
			pyro_winfo_parent, Tk_WinfoCmd, pyro_conv_rivetobj)
PYROARGCOMMAND("winfo","pathname",		\
			pyro_winfo_pathname, Tk_WinfoCmd, pyro_conv_rivetobj)
PYROARGCOMMAND("winfo","pixels",		\
			pyro_winfo_pixels, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","pointerx",		\
			pyro_winfo_pointerx, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","pointerxy",		\
			pyro_winfo_pointerxy, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","pointery",		\
			pyro_winfo_pointery, Tk_WinfoCmd, pyro_conv_tuple_int)
PYROARGCOMMAND("winfo","reqheight",		\
			pyro_winfo_reqheight, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","reqwidth",		\
			pyro_winfo_reqwidth, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","rgb",			\
			pyro_winfo_rgb, Tk_WinfoCmd, pyro_conv_tuple_int)
PYROARGCOMMAND("winfo","screen",		\
			pyro_winfo_screen, Tk_WinfoCmd, pyro_conv_str2)
PYROARGCOMMAND("winfo","screencells",		\
			pyro_winfo_screencells, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","screendepth",		\
			pyro_winfo_screendepth, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","screenheight",		\
			pyro_winfo_screenheight, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","screenmmheight",	\
			pyro_winfo_screenmmheight, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","screenmmwidth",		\
			pyro_winfo_screenmmwidth, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","screenvisual",		\
			pyro_winfo_screenvisual, Tk_WinfoCmd, pyro_conv_str2)
PYROARGCOMMAND("winfo","screenwidth",		\
			pyro_winfo_screenwidth, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","server",		\
			pyro_winfo_server, Tk_WinfoCmd, pyro_conv_str2)
PYROARGCOMMAND("winfo","toplevel",		\
			pyro_winfo_toplevel, Tk_WinfoCmd, pyro_conv_rivetobj)
PYROARGCOMMAND("winfo","viewable",		\
			pyro_winfo_viewable, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","visual",	\
			pyro_winfo_visual, Tk_WinfoCmd, pyro_conv_str2)
PYROARGCOMMAND("winfo","visualsavailable",	\
			pyro_winfo_visualsavailable, Tk_WinfoCmd, pyro_conv_str)
PYROARGCOMMAND("winfo","vrootheight",		\
			pyro_winfo_vrootheight, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","vrootwidth",		\
			pyro_winfo_vrootwidth, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","vrootx",		\
			pyro_winfo_vrootx, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","vrooty",		\
			pyro_winfo_vrooty, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","width",		\
			pyro_winfo_width, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","x",		\
			pyro_winfo_x, Tk_WinfoCmd, pyro_conv_int)
PYROARGCOMMAND("winfo","y",		\
			pyro_winfo_y, Tk_WinfoCmd, pyro_conv_int)


static void
pyro_init_winfo_ops(PyObject *dict)
{
	Pyro_ms ms;
	static Pyrometh methods[] =  {
		PYROMDECL("width", pyro_winfo_width),
		PYROMDECL("height", pyro_winfo_height),
		PYROMDECL("reqheight", pyro_winfo_reqheight),
		PYROMDECL("reqwidth", pyro_winfo_reqwidth),
		PYROMDECL("toplevel", pyro_winfo_toplevel),
		PYROMDECL("x", pyro_winfo_x),
		PYROMDECL("y", pyro_winfo_y),

		PYROMDECL("vrootheight", pyro_winfo_vrootheight),
		PYROMDECL("vrootwidth", pyro_winfo_vrootwidth),
		PYROMDECL("vrootx", pyro_winfo_vrootx),
		PYROMDECL("vrooty", pyro_winfo_vrooty),

		/* unsorted */
		PYROMDECL("atom", pyro_winfo_atom),
		PYROMDECL("atomname", pyro_winfo_atomname),
		PYROMDECL("cells", pyro_winfo_cells),
		PYROMDECL("children", pyro_winfo_children),
		PYROMDECL("_class", pyro_winfo_class),
		PYROMDECL("containing", pyro_winfo_containing),
		PYROMDECL("colormapfull", pyro_winfo_colormapfull),
		PYROMDECL("depth", pyro_winfo_depth),
		PYROMDECL("exists", pyro_winfo_exists),
		PYROMDECL("fpixels", pyro_winfo_fpixels),
		PYROMDECL("geometry", pyro_winfo_geometry),
		PYROMDECL("id", pyro_winfo_id),
		/* PYROMDECL("interps", pyro_winfo_interps),*/
		PYROMDECL("ismapped", pyro_winfo_ismapped),
		PYROMDECL("manager", pyro_winfo_manager),
		/* PYROMDECL("name", pyro_winfo_name), */
		PYROMDECL("parent", pyro_winfo_parent),
		PYROMDECL("pathname", pyro_winfo_pathname),
		PYROMDECL("pixels", pyro_winfo_pixels),
		PYROMDECL("pointerx", pyro_winfo_pointerx),
		PYROMDECL("pointerxy", pyro_winfo_pointerxy),
		PYROMDECL("pointery", pyro_winfo_pointery),
		PYROMDECL("rgb", pyro_winfo_rgb),
		PYROMDECL("screen", pyro_winfo_screen),
		PYROMDECL("screencells", pyro_winfo_screencells),
		PYROMDECL("screendepth", pyro_winfo_screendepth),
		PYROMDECL("screenheight", pyro_winfo_screenheight),
		PYROMDECL("screenmmheight", pyro_winfo_screenmmheight),
		PYROMDECL("screenmmwidth", pyro_winfo_screenmmwidth),
		PYROMDECL("screenvisual", pyro_winfo_screenvisual),
		PYROMDECL("screenwidth", pyro_winfo_screenwidth),
		PYROMDECL("server", pyro_winfo_server),
		PYROMDECL("viewable", pyro_winfo_viewable),
		PYROMDECL("visual", pyro_winfo_visual),
		PYROMDECL("visualsavailable", pyro_winfo_visualsavailable),
		{0}
	};

	ms = PyObject_NEW(Pyro_method_set_struct, &PyroMsType);
	ms->methods = pyro_init_methods(methods);
	ms->callfn = 0;

	PyDict_SetItemString(dict, "winfo", (PyObject *)ms);
}
/* end winfo */
/*----------------------------------------------------------------------*/
/* begin wm */

PYROARGCOMMAND("wm","aspect",	\
				pyro_wm_aspect, Tk_WmCmd, pyro_conv_tuple_int)
PYROARGCOMMAND("wm","client",	\
				pyro_wm_client, Tk_WmCmd, pyro_conv_str2)
PYROARGCOMMAND("wm","colormapwindows",	\
				pyro_wm_colormapwindows, Tk_WmCmd, pyro_conv_tuple_rivetobj)
PYROARGCOMMAND("wm","command",	\
				pyro_wm_command, Tk_WmCmd, pyro_conv_str2)
PYROARGCOMMAND("wm","deiconify",	\
				pyro_wm_deiconify, Tk_WmCmd, pyro_conv_none)
PYROARGCOMMAND("wm","focusmodel",	\
				pyro_wm_focusmodel, Tk_WmCmd, pyro_conv_str)
PYROARGCOMMAND("wm","frame",	\
				pyro_wm_frame, Tk_WmCmd, pyro_conv_str)
PYROARGCOMMAND("wm","geometry",	\
				pyro_wm_geometry, Tk_WmCmd, pyro_conv_str)
PYROARGCOMMAND("wm","grid",	\
				pyro_wm_grid, Tk_WmCmd, pyro_conv_tuple_int)
PYROARGCOMMAND("wm","group",	\
				pyro_wm_group, Tk_WmCmd, pyro_conv_rivetobj)
PYROARGCOMMAND("wm","iconbitmap",	\
				pyro_wm_iconbitmap, Tk_WmCmd, pyro_conv_str)
PYROARGCOMMAND("wm","iconify",	\
				pyro_wm_iconify, Tk_WmCmd, pyro_conv_none)
PYROARGCOMMAND("wm","iconmask",	\
				pyro_wm_iconmask, Tk_WmCmd, pyro_conv_str)
PYROARGCOMMAND("wm","iconname",	\
				pyro_wm_iconname, Tk_WmCmd, pyro_conv_str)
PYROARGCOMMAND("wm","iconposition",	\
				pyro_wm_iconposition, Tk_WmCmd, pyro_conv_tuple_int)
PYROARGCOMMAND("wm","iconwindow",	\
				pyro_wm_iconwindow, Tk_WmCmd, pyro_conv_rivetobj)
PYROARGCOMMAND("wm","maxsize",	\
				pyro_wm_maxsize, Tk_WmCmd, pyro_conv_tuple_int)
PYROARGCOMMAND("wm","minsize",	\
				pyro_wm_minsize, Tk_WmCmd, pyro_conv_tuple_int)
PYROARGCOMMAND("wm","overrideredirect",	\
				pyro_wm_overrideredirect, Tk_WmCmd, pyro_conv_int)
PYROARGCOMMAND("wm","positionfrom",	\
				pyro_wm_positionfrom, Tk_WmCmd, pyro_conv_str2)
PYROARGCOMMAND("wm","resizable",	\
				pyro_wm_resizable, Tk_WmCmd, pyro_conv_tuple_int)
PYROARGCOMMAND("wm","sizefrom",	\
				pyro_wm_sizefrom, Tk_WmCmd, pyro_conv_str)
PYROARGCOMMAND("wm","state",	\
				pyro_wm_state, Tk_WmCmd, pyro_conv_str2)
PYROARGCOMMAND("wm","title",	\
				pyro_wm_title, Tk_WmCmd, pyro_conv_str)
PYROARGCOMMAND("wm","tracing",	\
				pyro_wm_tracing, Tk_WmCmd, pyro_conv_str)
PYROARGCOMMAND("wm","transient",	\
				pyro_wm_transient, Tk_WmCmd, pyro_conv_rivetobj)
PYROARGCOMMAND("wm","withdraw",	\
				pyro_wm_withdraw, Tk_WmCmd, pyro_conv_none)

/* do wm protocol special, because of callbacks */
static PyObject *
pyro_wm_protocol(PyObject *po, PyObject *args, PyObject *d)
{
	PyObject *rv;

	rv = pyro_cmd("wm", "protocol", Tk_WmCmd, 0, args, d, pyro_conv_passthrough);
	if(!rv)
		return NULL;
	if(PyTuple_Size(args) == 1) {
		/* only window specified, return tuple of strings for
		 * each of the protocols with handlers. */
		return pyro_conv_tuple_str((char*)rv, args);
	}
	if(PyTuple_Size(args) == 2) {
		/* a window and a name specified, return the callback */
		Callback cb;
		cb = (Callback) rv;
		if(cb && cb->type == PYRO_UNIQUE_TOKEN)
			/* it is incref'd in tkwm code */
			return (PyObject *)(cb->assoc);
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static void
pyro_init_wm_ops(PyObject *dict)
{
	Pyro_ms ms;
	static Pyrometh methods[] =  {
		PYROMDECL("aspect", pyro_wm_aspect),
		PYROMDECL("client", pyro_wm_client),
		PYROMDECL("colormapwindows", pyro_wm_colormapwindows),
		PYROMDECL("command", pyro_wm_command),
		PYROMDECL("deiconify", pyro_wm_deiconify),
		PYROMDECL("focusmodel", pyro_wm_focusmodel),
		PYROMDECL("frame", pyro_wm_frame),
		PYROMDECL("geometry", pyro_wm_geometry),
		PYROMDECL("grid", pyro_wm_grid),
		PYROMDECL("group", pyro_wm_group),
		PYROMDECL("iconbitmap", pyro_wm_iconbitmap),
		PYROMDECL("iconify", pyro_wm_iconify),
		PYROMDECL("iconmask", pyro_wm_iconmask),
		PYROMDECL("iconname", pyro_wm_iconname),
		PYROMDECL("iconposition", pyro_wm_iconposition),
		PYROMDECL("iconwindow", pyro_wm_iconwindow),
		PYROMDECL("maxsize", pyro_wm_maxsize),
		PYROMDECL("minsize", pyro_wm_minsize),
		PYROMDECL("overrideredirect", pyro_wm_overrideredirect),
		PYROMDECL("positionfrom", pyro_wm_positionfrom),
		PYROMDECL("protocol", pyro_wm_protocol),
		PYROMDECL("resizable", pyro_wm_resizable),
		PYROMDECL("sizefrom", pyro_wm_sizefrom),
		PYROMDECL("state", pyro_wm_state),
		PYROMDECL("title", pyro_wm_title),
		PYROMDECL("tracing", pyro_wm_tracing),
		PYROMDECL("transient", pyro_wm_transient),
		PYROMDECL("withdraw", pyro_wm_withdraw),
		{0}
	};

	ms = PyObject_NEW(Pyro_method_set_struct, &PyroMsType);
	ms->methods = pyro_init_methods(methods);
	ms->callfn = 0;

	PyDict_SetItemString(dict, "wm", (PyObject *)ms);
}
/* end wm */
/*----------------------------------------------------------------------*/

static PyObject *
pyro_call_tkmethod(Pyroobj po, PyObject *args, PyObject *d,
	char *tkarg1, char *tkarg2,
	PyObject *(*convertfunc)())
{
	Argcv argcv;
	Rivetobj ro;
	int rv;

	if(!po || !is_pyroobj(po))
		return NULL;

	ro = po->rivetobj;
	if(tkarg2) {
		INIT_ARGCV(argcv, 3)
		argcv.argv[0] = rivet_path(ro);
		argcv.argv[1] = tkarg1;
		argcv.argv[2] = tkarg2;
	} else {
		INIT_ARGCV(argcv, 2)
		argcv.argv[0] = rivet_path(ro);
		argcv.argv[1] = tkarg1;
	}

	if (pyro_parse_argcv(&argcv, args))
		return NULL;
	if (d && pyro_parse_argcv(&argcv, d))
		return NULL;

	Tcl_SetResult(ro->interp, 0, 0);
	rv = (ro->rivet_class->widgetcmd) (ro, ro->interp,
			argcv.argc, argcv.argv);
	FREE_ARGCV(argcv)
	if (rv == TCL_ERROR) {
		PyErr_SetString(pyro_error, ro->interp->result);
		return NULL;
	}
	return (convertfunc)(ro->interp->result, args);
}

#define OBJECTMETHODCMD( _tkmethodname, _methodname, _convert)		\
	static PyObject *						\
	_methodname (PyObject *po, PyObject *args, PyObject *d)		\
	{								\
		return pyro_call_tkmethod((Pyroobj)po, args, d,		\
				_tkmethodname, 0, _convert);		\
	}

#define OBJECTMETHODCMD2( _tkmethodname, _tkmethodname2, _methodname, _convert)\
	static PyObject *						\
	_methodname (PyObject *po, PyObject *args, PyObject *d)		\
	{								\
		return pyro_call_tkmethod((Pyroobj)po, args, d,		\
				_tkmethodname, _tkmethodname2, _convert);\
	}


#define CREATE_ANY_CHILD \
	PYROMDECL("button",	pyro_button),\
	PYROMDECL("canvas",	pyro_canvas),\
	PYROMDECL("checkbutton",pyro_checkbutton),\
	PYROMDECL("entry",	pyro_entry),\
	PYROMDECL("frame",	pyro_frame),\
	PYROMDECL("label",	pyro_label),\
	PYROMDECL("listbox",	pyro_listbox),\
	PYROMDECL("menu",	pyro_menu),\
	PYROMDECL("menubutton",	pyro_menubutton),\
	PYROMDECL("message",	pyro_message),\
	PYROMDECL("radiobutton",pyro_radiobutton),\
	PYROMDECL("scale",	pyro_scale),\
	PYROMDECL("scrollbar",	pyro_scrollbar),\
	PYROMDECL("text",	pyro_text),\
	PYROMDECL("toplevel",	pyro_toplevel),

#define CREATE_TOPLEVEL_CHILD \
	PYROMDECL("menu",	pyro_menu),\
	PYROMDECL("toplevel",	pyro_toplevel),

#define GEOMETRY_METHODDEF \
	PYROMDECL("pack",	pyro_pack), \
	PYROMDECL("place",	pyro_place),

#define GENERIC_METHODDEF \
	PYROMDECL("cget",	pyro_m_cget), \
	PYROMDECL("configure",	pyro_m_configure), \
	PYROMDECL2("bind",	pyro_bind), \
	PYROMDECL2("parent",	pyro_parent), \
	PYROMDECL2("children",	pyro_children), \
	PYROMDECL2("destroy",	pyro_destroy_object), \
	PYROMDECL2("name",	pyro_name),


/* These are shared among two or more widgets */
OBJECTMETHODCMD("activate",	pyro_m_activate, pyro_conv_none)
OBJECTMETHODCMD("bbox",		pyro_m_bbox, pyro_conv_tuple_int)
OBJECTMETHODCMD("cget",		pyro_m_cget, pyro_conv_cget)
OBJECTMETHODCMD("configure",	pyro_m_configure, pyro_conv_configure)
OBJECTMETHODCMD("coords",	pyro_m_coords, pyro_conv_tuple_int)
OBJECTMETHODCMD("delete",	pyro_m_delete, pyro_conv_none)
OBJECTMETHODCMD("get",		pyro_m_get, pyro_conv_str2)
OBJECTMETHODCMD("icursor",	pyro_m_icursor, pyro_conv_none)
OBJECTMETHODCMD("identify",	pyro_m_identify, pyro_conv_str)
OBJECTMETHODCMD("index",	pyro_m_index, pyro_conv_int)
OBJECTMETHODCMD("index",	pyro_m_index_str, pyro_conv_str)
OBJECTMETHODCMD("insert",	pyro_m_insert, pyro_conv_none)
OBJECTMETHODCMD2("scan","dragto",	pyro_m_scan_dragto, pyro_conv_none)
OBJECTMETHODCMD2("scan","mark",		pyro_m_scan_mark, pyro_conv_none)
OBJECTMETHODCMD("see",		pyro_m_see, pyro_conv_none)
OBJECTMETHODCMD2("selection","clear",	pyro_m_selection_clear, pyro_conv_none)
OBJECTMETHODCMD("set",		pyro_m_set, pyro_conv_none)
OBJECTMETHODCMD("type",		pyro_m_type, pyro_conv_str)
OBJECTMETHODCMD("xview",	pyro_m_xview, pyro_conv_tuple_float)
OBJECTMETHODCMD("yview",	pyro_m_yview, pyro_conv_tuple_float)


/*
 * For those widgets with only a configure and cget.
 * label, frame, menubutton, message
 */
static Pyrometh simple_widget_mdef[] = {
	GENERIC_METHODDEF
	GEOMETRY_METHODDEF
	CREATE_ANY_CHILD
	{0},
};

static Pyrometh toplevel_widget_mdef[] = {
	GENERIC_METHODDEF
	CREATE_ANY_CHILD
	{0},
};


/* button */
OBJECTMETHODCMD("deselect",	pyro_m_deselect, pyro_conv_none)
OBJECTMETHODCMD("flash",	pyro_m_flash, pyro_conv_none)
OBJECTMETHODCMD("invoke",	pyro_m_invoke_button, pyro_conv_none)
OBJECTMETHODCMD("select",	pyro_m_select, pyro_conv_none)
OBJECTMETHODCMD("toggle",	pyro_m_toggle, pyro_conv_none)

static Pyrometh button_widget_mdef[] = {
	GENERIC_METHODDEF
	GEOMETRY_METHODDEF
	PYROMDECL("flash",	pyro_m_flash),
	PYROMDECL("invoke",	pyro_m_invoke_button),
	CREATE_TOPLEVEL_CHILD
	{0},
};

static Pyrometh cbutton_widget_mdef[] = {
	GENERIC_METHODDEF
	GEOMETRY_METHODDEF
	PYROMDECL("deselect",	pyro_m_deselect),
	PYROMDECL("select",	pyro_m_select),
	PYROMDECL("flash",	pyro_m_flash),
	PYROMDECL("invoke",	pyro_m_invoke_button),
	PYROMDECL("toggle",	pyro_m_toggle),
	CREATE_TOPLEVEL_CHILD
	{0},
};

static Pyrometh rbutton_widget_mdef[] = {
	GENERIC_METHODDEF
	GEOMETRY_METHODDEF
	PYROMDECL("deselect",	pyro_m_deselect),
	PYROMDECL("select",	pyro_m_select),
	PYROMDECL("flash",	pyro_m_flash),
	PYROMDECL("invoke",	pyro_m_invoke_button),
	CREATE_TOPLEVEL_CHILD
	{0},
};

OBJECTMETHODCMD2("selection","adjust",	pyro_m_selection_adjust, pyro_conv_none)
OBJECTMETHODCMD2("selection","from",	pyro_m_selection_from, pyro_conv_none)
OBJECTMETHODCMD2("selection","present",	pyro_m_selection_present, pyro_conv_int)
OBJECTMETHODCMD2("selection","range",	pyro_m_selection_range, pyro_conv_none)
OBJECTMETHODCMD2("selection","to",	pyro_m_selection_to, pyro_conv_none)

static Pyrometh entry_widget_mdef[] = {
	GENERIC_METHODDEF
	GEOMETRY_METHODDEF
	PYROMDECL("delete",	pyro_m_delete),
	PYROMDECL("get",	pyro_m_get),
	PYROMDECL("icursor",	pyro_m_icursor),
	PYROMDECL("index",	pyro_m_index),
	PYROMDECL("insert",	pyro_m_insert),
	PYROMDECL("scan_dragto",pyro_m_scan_dragto),
	PYROMDECL("scan_mark",	pyro_m_scan_mark),
	PYROMDECL("selection_adjust",pyro_m_selection_adjust),
	PYROMDECL("selection_clear",pyro_m_selection_clear),
	PYROMDECL("selection_from",pyro_m_selection_from),
	PYROMDECL("selection_present",pyro_m_selection_present),
	PYROMDECL("selection_range",pyro_m_selection_range),
	PYROMDECL("selection_to",pyro_m_selection_to),
	PYROMDECL("xview",	pyro_m_xview),
	CREATE_TOPLEVEL_CHILD
	{0},
};

OBJECTMETHODCMD("get",		pyro_m_get_scale, pyro_conv_float)

static Pyrometh scale_widget_mdef[] = {
	GENERIC_METHODDEF
	GEOMETRY_METHODDEF
        PYROMDECL("coords",	pyro_m_coords),
	PYROMDECL("get",	pyro_m_get_scale),
	PYROMDECL("identify",	pyro_m_identify),
	PYROMDECL("set",	pyro_m_set),
	CREATE_TOPLEVEL_CHILD
	{0},
};

OBJECTMETHODCMD("addtag",	pyro_m_addtag, pyro_conv_none)
OBJECTMETHODCMD("canvasx",	pyro_m_canvasx, pyro_conv_float)
OBJECTMETHODCMD("canvasy",	pyro_m_canvasy, pyro_conv_float)
OBJECTMETHODCMD("create",	pyro_m_create, pyro_conv_str)
OBJECTMETHODCMD("dchars",	pyro_m_dchars, pyro_conv_none)
OBJECTMETHODCMD("dtag",		pyro_m_dtag, pyro_conv_none)
OBJECTMETHODCMD("find",		pyro_m_find, pyro_conv_tuple_str)
OBJECTMETHODCMD("focus",	pyro_m_focus, pyro_conv_int_none)
OBJECTMETHODCMD("gettags",	pyro_m_gettags, pyro_conv_tuple_str)
OBJECTMETHODCMD("itemcget",	pyro_m_itemcget, pyro_conv_cget)
OBJECTMETHODCMD("itemconfigure",pyro_m_itemconfigure, pyro_conv_configure)
OBJECTMETHODCMD("lower",	pyro_m_lower, pyro_conv_none)
OBJECTMETHODCMD("move",		pyro_m_move, pyro_conv_none)
OBJECTMETHODCMD("postscript",	pyro_m_postscript, pyro_conv_str2)
OBJECTMETHODCMD("raise",	pyro_m_raise, pyro_conv_none)
OBJECTMETHODCMD("scale",	pyro_m_scale, pyro_conv_none)
OBJECTMETHODCMD2("select","adjust",	pyro_m_select_adjust, pyro_conv_none)
OBJECTMETHODCMD2("select","clear",	pyro_m_select_clear, pyro_conv_none)
OBJECTMETHODCMD2("select","from",	pyro_m_select_from, pyro_conv_none)
OBJECTMETHODCMD2("select","item",	pyro_m_select_item, pyro_conv_str2)
OBJECTMETHODCMD2("select","to",		pyro_m_select_to, pyro_conv_none)

static Pyrometh canvas_widget_mdef[] = {
	GENERIC_METHODDEF
	PYROMDECL("addtag",	pyro_m_addtag),
	PYROMDECL("bbox",	pyro_m_bbox),
	PYROMDECL("itemcget",	pyro_m_itemcget),
	PYROMDECL("itemconfigure",pyro_m_itemconfigure),
	PYROMDECL2("tag_bind",	pyro_tag_bind),
	PYROMDECL("canvasx",	pyro_m_canvasx),
	PYROMDECL("canvasy",	pyro_m_canvasy),
        PYROMDECL("coords",     pyro_m_coords),
	PYROMDECL("create",     pyro_m_create),
  	PYROMDECL("dchars",     pyro_m_dchars),
	PYROMDECL("delete",	pyro_m_delete),
	PYROMDECL("dtag",	pyro_m_dtag),
	PYROMDECL("find",	pyro_m_find),
	PYROMDECL("focus",	pyro_m_focus),
	PYROMDECL("gettags",	pyro_m_gettags),
	PYROMDECL("icursor",	pyro_m_icursor),
	PYROMDECL("index",	pyro_m_index),
	PYROMDECL("insert",	pyro_m_insert),
	PYROMDECL("lower",	pyro_m_lower),
	PYROMDECL("move",	pyro_m_move),
	PYROMDECL("postscript",	pyro_m_postscript),

	PYROMDECL("raise",	pyro_m_raise),
	/* Note there is a method name conflict with scale (create a scale
	 * child vs. the canvas's scale method), so handle this
	 * by making sure this is before the "create a scale" method,
	 * thus it will be found first in the search.
	 */
	PYROMDECL("scale",	pyro_m_scale),

	PYROMDECL("scan_dragto",pyro_m_scan_dragto),
	PYROMDECL("scan_mark",	pyro_m_scan_mark),

	PYROMDECL("select_adjust",pyro_m_select_adjust),
	PYROMDECL("select_clear",pyro_m_select_clear),
	PYROMDECL("select_from", pyro_m_select_from),
	PYROMDECL("select_item", pyro_m_select_item),
	PYROMDECL("select_to",   pyro_m_select_to),

	PYROMDECL("type",	pyro_m_type),
	PYROMDECL("xview",	pyro_m_xview),
	PYROMDECL("yview",	pyro_m_yview),
	GEOMETRY_METHODDEF
	CREATE_ANY_CHILD
	{0},
};

OBJECTMETHODCMD("curselection",	pyro_m_curselection, pyro_conv_tuple_int)
OBJECTMETHODCMD("get",		pyro_m_get_listbox, pyro_conv_str_tuple_str)
OBJECTMETHODCMD("nearest",	pyro_m_nearest, pyro_conv_int)
OBJECTMETHODCMD("size",		pyro_m_size, pyro_conv_int)
OBJECTMETHODCMD2("selection","anchor",	pyro_m_selection_anchor, pyro_conv_none)
OBJECTMETHODCMD2("selection","includes",pyro_m_selection_includes,pyro_conv_int)
OBJECTMETHODCMD2("selection","set",	pyro_m_selection_set, pyro_conv_none)

static Pyrometh listbox_widget_mdef[] = {
	GENERIC_METHODDEF
	GEOMETRY_METHODDEF
	PYROMDECL("activate",	pyro_m_activate),
	PYROMDECL("bbox",	pyro_m_bbox),
	PYROMDECL("curselection",pyro_m_curselection),
	PYROMDECL("delete",	pyro_m_delete),
	PYROMDECL("get",	pyro_m_get_listbox),
	PYROMDECL("index",	pyro_m_index),
	PYROMDECL("insert",	pyro_m_insert),
	PYROMDECL("nearest",	pyro_m_nearest),
	PYROMDECL("scan_dragto",pyro_m_scan_dragto),
	PYROMDECL("scan_mark",	pyro_m_scan_mark),
	PYROMDECL("see",	pyro_m_see),
	PYROMDECL("selection_anchor",	pyro_m_selection_anchor),
	PYROMDECL("selection_clear",	pyro_m_selection_clear),
	PYROMDECL("selection_includes",	pyro_m_selection_includes),
	PYROMDECL("selection_set",	pyro_m_selection_set),
	PYROMDECL("size",	pyro_m_size),
	PYROMDECL("xview",	pyro_m_xview),
	PYROMDECL("yview",	pyro_m_yview),
	CREATE_TOPLEVEL_CHILD
	{0},
};

OBJECTMETHODCMD("entrycget",	pyro_m_entrycget, pyro_conv_cget)
OBJECTMETHODCMD("entryconfigure",pyro_m_entryconfigure, pyro_conv_configure)
OBJECTMETHODCMD("post",		pyro_m_post, pyro_conv_none) /* gack, the postcommand return value or "" is returned */
OBJECTMETHODCMD("postcascade",	pyro_m_postcascade, pyro_conv_none) /* gack, don't know return value */
OBJECTMETHODCMD("unpost",	pyro_m_unpost, pyro_conv_none)
OBJECTMETHODCMD("yposition",	pyro_m_yposition, pyro_conv_int)

OBJECTMETHODCMD("add",		pyro_m_add_mentry, pyro_conv_none)
OBJECTMETHODCMD("insert",	pyro_m_insert_mentry, pyro_conv_none)
OBJECTMETHODCMD("invoke",	pyro_m_invoke_menu, pyro_conv_none) /* gack return value of invoke'd */

static Pyrometh menu_widget_mdef[] = {
	GENERIC_METHODDEF
        PYROMDECL("activate",   pyro_m_activate),
	PYROMDECL("add", 	pyro_m_add_mentry),
	PYROMDECL("delete",	pyro_m_delete),
	PYROMDECL("entrycget",	pyro_m_entrycget),
	PYROMDECL("entryconfigure",pyro_m_entryconfigure),
	PYROMDECL("index",	pyro_m_index_str),
	PYROMDECL("insert",	pyro_m_insert_mentry),
	PYROMDECL("invoke",	pyro_m_invoke_menu),
	PYROMDECL("post",	pyro_m_post),
	PYROMDECL("postcascade",pyro_m_postcascade),
	PYROMDECL("type",	pyro_m_type),
	PYROMDECL("unpost",	pyro_m_unpost),
	PYROMDECL("yposition",	pyro_m_yposition),
	CREATE_TOPLEVEL_CHILD
	{0},
};

OBJECTMETHODCMD("fraction",	pyro_m_fraction, pyro_conv_float)
OBJECTMETHODCMD("delta",	pyro_m_delta, pyro_conv_float)
OBJECTMETHODCMD("get",		pyro_m_get_sbar, pyro_conv_tuple_float)
OBJECTMETHODCMD("activate",	pyro_m_activate_sbar, pyro_conv_str)

static Pyrometh scrollbar_widget_mdef[] = {
	GENERIC_METHODDEF
	PYROMDECL("set",	pyro_m_set),
	GEOMETRY_METHODDEF
        PYROMDECL("activate",   pyro_m_activate_sbar),
	PYROMDECL("delta",	pyro_m_delta),
	PYROMDECL("fraction",	pyro_m_fraction),
	PYROMDECL("get",	pyro_m_get_sbar),
	PYROMDECL("identify",	pyro_m_identify),
	CREATE_TOPLEVEL_CHILD
	{0},
};

OBJECTMETHODCMD("compare",		pyro_m_compare, pyro_conv_int)
OBJECTMETHODCMD("debug",		pyro_m_debug, pyro_conv_str2)
OBJECTMETHODCMD("dlineinfo",		pyro_m_dlineinfo, pyro_conv_tuple_int)
OBJECTMETHODCMD2("mark","gravity",	pyro_m_mark_gravity, pyro_conv_str2)
OBJECTMETHODCMD2("mark","names",	pyro_m_mark_names, pyro_conv_tuple_str)
OBJECTMETHODCMD2("mark","set",		pyro_m_mark_set, pyro_conv_none)
OBJECTMETHODCMD2("mark","unset",	pyro_m_mark_unset, pyro_conv_none)
OBJECTMETHODCMD("search",		pyro_m_search, pyro_conv_str2)
OBJECTMETHODCMD2("tag","add",		pyro_m_tag_add, pyro_conv_none)
OBJECTMETHODCMD2("tag","cget",		pyro_m_text_tag_cget, pyro_conv_cget)
OBJECTMETHODCMD2("tag","configure",	pyro_m_text_tag_configure, pyro_conv_configure)
OBJECTMETHODCMD2("tag","delete",	pyro_m_text_tag_delete, pyro_conv_none)
OBJECTMETHODCMD2("tag","lower",		pyro_m_text_tag_lower, pyro_conv_none)
OBJECTMETHODCMD2("tag","names",		pyro_m_text_tag_names, pyro_conv_tuple_str)
OBJECTMETHODCMD2("tag","nextrange",	pyro_m_text_tag_nextrange, pyro_conv_tuple_str)
OBJECTMETHODCMD2("tag","raise",		pyro_m_text_tag_raise, pyro_conv_none)
OBJECTMETHODCMD2("tag","ranges",	pyro_m_text_tag_ranges, pyro_conv_tuple_str)
OBJECTMETHODCMD2("tag","remove",	pyro_m_text_tag_remove, pyro_conv_none)
OBJECTMETHODCMD2("window","cget",	pyro_m_window_cget, pyro_conv_cget)
OBJECTMETHODCMD2("window","configure",	pyro_m_window_configure, pyro_conv_configure)
OBJECTMETHODCMD2("window","create",	pyro_m_window_create, pyro_conv_none)
OBJECTMETHODCMD2("window","names",	pyro_m_window_names, pyro_conv_tuple_rivetobj)


static Pyrometh text_widget_mdef[] = {
	GENERIC_METHODDEF
	PYROMDECL("tag_add",	pyro_m_tag_add),
	PYROMDECL("tag_configure",pyro_m_text_tag_configure),
	PYROMDECL2("tag_bind",	pyro_tag_bind),
	PYROMDECL("tag_cget",	pyro_m_text_tag_cget),
	PYROMDECL("tag_delete",	pyro_m_text_tag_delete),
	PYROMDECL("tag_lower",	pyro_m_text_tag_lower),
	PYROMDECL("tag_names",	pyro_m_text_tag_names),
	PYROMDECL("tag_nextrange",pyro_m_text_tag_nextrange),
	PYROMDECL("tag_raise",	pyro_m_text_tag_raise),
	PYROMDECL("tag_ranges", pyro_m_text_tag_ranges),
	PYROMDECL("tag_remove", pyro_m_text_tag_remove),
	PYROMDECL("window_create",pyro_m_window_create),

	/* unsorted */
	GEOMETRY_METHODDEF
	PYROMDECL("bbox",	pyro_m_bbox),
	PYROMDECL("compare",	pyro_m_compare),
        PYROMDECL("debug",      pyro_m_debug),
	PYROMDECL("delete",	pyro_m_delete),
	PYROMDECL("dlineinfo",  pyro_m_dlineinfo),
	PYROMDECL("get",	pyro_m_get),
	PYROMDECL("index",	pyro_m_index_str),
	PYROMDECL("insert",	pyro_m_insert),

	PYROMDECL("mark_gravity",pyro_m_mark_gravity),
	PYROMDECL("mark_names",	pyro_m_mark_names),
	PYROMDECL("mark_set",	pyro_m_mark_set),
	PYROMDECL("mark_unset",	pyro_m_mark_unset),

	PYROMDECL("scan_mark",	pyro_m_scan_mark),
	PYROMDECL("scan_dragto",pyro_m_scan_dragto),

	PYROMDECL("search",	pyro_m_search),
	PYROMDECL("see",	pyro_m_see),


	PYROMDECL("window_cget",pyro_m_window_cget),
	PYROMDECL("window_configure",pyro_m_window_configure),
	PYROMDECL("window_names",pyro_m_window_names),

	PYROMDECL("xview",	pyro_m_xview),
	PYROMDECL("yview",	pyro_m_yview),
	CREATE_ANY_CHILD
	{0},
};

/*----------------------------------------------------------------------*/

static void
pyroobj_dealloc(Pyroobj p)
{
	if(p->rivetobj)
		rivet_destroy_window(p->rivetobj);
	if(p->dictionary)
		Py_DECREF(p->dictionary);
	p->rivetobj = 0;
	p->dictionary = 0;
}

void
pyro_destroy_notify(Rivetobj ro)
{
	/*
	 * The rivetobj is going away, flag this fact inside the pyroobj by
	 * setting pointer to it to NULL.
	 * The python object lives on until garbage collected.
	 * Children will be cleaned up recursively by tk.
	 */
	Pyroobj po;
	po = (Pyroobj) ro->language_object;
	if(!po)
		return;
	po->rivetobj = 0;
	if(po->dictionary) {
		Py_DECREF(po->dictionary);
		po->dictionary = 0;
	}
	Py_DECREF(po); /* decrement for the parent's reference */
}

static void
pyro_callback_incref(Callback cb)
{
	if(cb->type == PYRO_UNIQUE_TOKEN)
		Py_INCREF((PyObject*)cb->assoc);
}
static void
pyro_callback_decref(Callback cb)
{
	if(cb->type == PYRO_UNIQUE_TOKEN)
		Py_DECREF((PyObject*)cb->assoc);
}

static PyObject *pyro_save_thread;
static void
pyro_begin_allow_threads()
{
#ifdef WITH_THREAD
	Py_BEGIN_ALLOW_THREADS
	pyro_save_thread = _save;
	} /* macro does not close bracket */
#endif
}

static void
pyro_end_allow_threads()
{
#ifdef WITH_THREAD
	PyObject *_save = pyro_save_thread;
	{ /* macro does not open bracket */
	Py_END_ALLOW_THREADS
#endif
}

static PyObject *
pyro_init(PyObject *self, PyObject *args, PyObject *d)
{
	extern char *getprogramname();
	Pyroobj po;
	Rivetobj ro;
	Argcv argcv;
	static short once;

	if(once == 1) {
		/* tk isn't clever about closing down, so prevent restarting
		 * after initialized once.  The only solution to this
		 * problem is to unload the rivetmodule and then import.
		 */
		PyErr_SetString(pyro_error, "Cannot initialize twice.");
		return NULL;
	}

	MainWindowClass->language_class_object = (ClientData) simple_widget_methods;

	po = PyObject_NEW(Pyro_object_struct, &PyroType);
	if(!po) {
		PyErr_SetString(pyro_error, "Could not create rivet object.");
		return NULL;
	}
	po->rivetobj = 0;

	INIT_ARGCV(argcv, 1)
	argcv.argv[0] = getprogramname();

	if(pyro_parse_argcv(&argcv, args)) {
		Py_DECREF(po);
		return NULL;
	}
	if(d && pyro_parse_argcv(&argcv, d)) {
		Py_DECREF(po);
		return NULL;
	}

	ro = rivet_init(argcv.argc, argcv.argv);
	FREE_ARGCV(argcv)

	if(!ro) {
		Py_DECREF(po);
		PyErr_SetString(pyro_error, ro->interp->result);
		return NULL;
	}
	once = 1;

	po->rivetobj = ro;
	ro->language_object = (ClientData) po;
	po->dictionary = 0;
	Py_INCREF(po); /* a ref for the parent to child relationship */

        rivet_notify.rivetobj_destroy_notify = pyro_destroy_notify;
        rivet_notify.decref_assoc_data = pyro_callback_decref;
        rivet_notify.incref_assoc_data = pyro_callback_incref;
        rivet_notify.begin_allow_threads = pyro_begin_allow_threads;
        rivet_notify.end_allow_threads = pyro_end_allow_threads;

	return (PyObject*) po;
}

static PyObject *
pyro_attrget(PyObject *unused, PyObject *args)
{
	/* This function allows python class definitions to
	 * supply a __getattr__ function so they can "inherit"
	 * widget methods such as cget, etc. just by doing a 
	 * __getattr__ = rivet.__getattr__.
	 */

	PyObject *obj;
	char *name;
	Pyroobj po;
	Pyrometh *methods;
	PyObject *rv0, *rv;

	if(PyArg_ParseTuple(args, "Os", &obj, &name) == NULL)
		return NULL;
	po = pyro_get_widget((Pyroobj)obj);
	if(!po)
		return NULL;
	methods=(Pyrometh*)(po->rivetobj->rivet_class->language_class_object);
	rv = pyro_find_method(methods, po, name);
	if(rv)
		return rv;
	PyErr_SetString(PyExc_AttributeError, name);
	return NULL;
}

static PyObject *
pyroobj_getattr(Pyroobj po, char *name)
{
	Pyrometh *methods;
	PyObject *rv0;
	PyObject *rv;

	/* assert(is_pyroobj(po); */
	if(!po->rivetobj) {
		PyErr_SetString(pyro_error, "Rivet object has been destroyed");
		return NULL;
	}
	methods=(Pyrometh*)(po->rivetobj->rivet_class->language_class_object);
	rv = pyro_find_method(methods, po, name);
	if(rv)
		return rv;

	PyErr_Clear();

	if(name[0] == '_') {
		if(!strcmp(name, "__dict__")) {
			if(!po->dictionary)
				po->dictionary = PyDict_New();
			rv = po->dictionary;
			goto done;
		}
		if(!strcmp(name, "__widget")) {
			rv = (PyObject *)po;
			goto done;
		}
	}
	if(po->dictionary) {
		rv = PyDict_GetItemString(po->dictionary, name);
		if(rv)
			goto done;
	}
	PyErr_SetString(PyExc_AttributeError, name);
	return NULL;

	done:
	if(rv)
		Py_INCREF(rv);
	return rv;
}

static int
pyroobj_setattr(Pyroobj po, char *name, PyObject *value)
{
	int rv;

	if(!po->dictionary) {
		po->dictionary = PyDict_New();
		if(!po->dictionary)
			return -1;
	}
	if(!value) {
		rv = PyDict_DelItemString(po->dictionary, name);
		if(rv < 0)
			PyErr_SetString(PyExc_AttributeError,
				"delete non-existing attribute");
		return rv;
        } else
                return PyDict_SetItemString(po->dictionary, name, value);
}

#ifdef DEBUG
static PyObject *
pyro_repr(Pyroobj po)
{
	char buf[60];
	sprintf(buf, "<rivetobj object refcnt=%d name=%s at 0x%x>",
		po->ob_refcnt, po->rivetobj? rivet_path(po->rivetobj):"", po);
	return PyString_FromString(buf);
}
#endif

static PyTypeObject PyroType = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,				/* size */
	"rivetobj",			/* name */
	sizeof(Pyro_object_struct),	/* size */
	0,				/* itemsize */
	(destructor)pyroobj_dealloc, 	/* dealloc */
	0,				/* print */
	(getattrfunc)pyroobj_getattr,	/* getattr */
	(setattrfunc)pyroobj_setattr,	/* setattr */
	0,				/* compare */
#ifdef DEBUG
	(reprfunc)pyro_repr,		/* repr */
#else
	(reprfunc)0,		/* repr */
#endif
	0,				/* as number */
	0,				/* as sequence */
	0, 				/* as mapping */
 	0,				/* tp_hash */
 	(ternaryfunc)0,			/* tp_call */
 	0,				/* tp_str */
};

/*----------------------------------------------------------------------*/
static void
pyro_cleanup()
{
	/* use Tk_DestroyWindow or rivet_destroy_window */
}


void
initrivet()
{
	static int init;
	PyObject *mod;
	PyObject *dict;
	Pyrometh *ptr;
	static Pyrometh newwidgets[] = {
		CREATE_ANY_CHILD
		{0}
	};
	static PyMethodDef module_methods[] = {
		{"init",	(PyCFunction)pyro_init, METH_VARARGS|METH_KEYWORDS},
		{"bindtags",	(PyCFunction)pyro_bindtags,1},
		{"bind",	(PyCFunction)pyro_bind, 1},
		{"destroy",	(PyCFunction)pyro_destroy_object, 1},
		{"lower",	(PyCFunction)pyro_lower, 1},
		{"raise",	(PyCFunction)pyro_raise, 1},
		{"update",	(PyCFunction)pyro_update, 1},
		{"tk_appname",	(PyCFunction)pyro_tk_appname, 1},
		{"after",	(PyCFunction)pyro_after, 1},
		{"bell",	(PyCFunction)pyro_bell, 1},
		{"fileevent",	(PyCFunction)pyro_fileevent, 1},
		{"focus",	(PyCFunction)pyro_focus, 1},
		{"mainloop",	(PyCFunction)pyro_mainloop, 1},
		{"mainwidget",	(PyCFunction)pyro_main_widget, 1},
		{"__getattr__",	(PyCFunction)pyro_attrget, 1},
		{"getvar",	(PyCFunction)pyro_getvar, 1},
		{"setvar",	(PyCFunction)pyro_setvar, 1},
		{"unsetvar",	(PyCFunction)pyro_unsetvar, 1},
		{"tracevar",	(PyCFunction)pyro_tracevar, 1},
		{"untracevar",	(PyCFunction)pyro_untracevar, 1},

		/******* to do: (possibly in rivet_util.py)
		tk_dialog, tk_focusNext, tk_focusPrev,
		tk_focusContinue, tk_focusFollowsMouse, 
		tk_popup, tkerror, tk_strictMotif
		*******/
		
		{NULL},
	};

	mod = Py_InitModule4("rivet", module_methods,
		"Rivet: python module interface to tk.",
		(PyObject *)NULL,
		PYTHON_API_VERSION);
	dict = PyModule_GetDict(mod);

	ptr = newwidgets;
	while(ptr->ml_name) {
		PyDict_SetItemString(dict, ptr->ml_name,
			PyCFunction_New(ptr, (PyObject*)0));
		ptr++;
	}

	pyro_error = PyString_FromString("rivet.error");
	PyDict_SetItemString(dict, "error", pyro_error);

	/* initialize the "command" methods ... */
	pyro_init_image_ops(dict);
	pyro_init_clipboard_ops(dict);
	pyro_init_grab_ops(dict);
	pyro_init_option_ops(dict);
	pyro_init_pack_ops(dict);
	pyro_init_place_ops(dict);
	pyro_init_selection_ops(dict);
	pyro_init_tkwait_ops(dict);
	pyro_init_winfo_ops(dict);
	pyro_init_wm_ops(dict);

	/* ... initialize the widget methods ... */
	simple_widget_methods = pyro_init_methods(simple_widget_mdef);
	toplevel_widget_methods = pyro_init_methods(toplevel_widget_mdef);
	button_widget_methods = pyro_init_methods(button_widget_mdef);
	cbutton_widget_methods = pyro_init_methods(cbutton_widget_mdef);
	rbutton_widget_methods = pyro_init_methods(rbutton_widget_mdef);
	entry_widget_methods = pyro_init_methods(entry_widget_mdef);
	scale_widget_methods = pyro_init_methods(scale_widget_mdef);
	canvas_widget_methods = pyro_init_methods(canvas_widget_mdef);
	listbox_widget_methods = pyro_init_methods(listbox_widget_mdef);
	menu_widget_methods = pyro_init_methods(menu_widget_mdef);
	scrollbar_widget_methods = pyro_init_methods(scrollbar_widget_mdef);
	text_widget_methods = pyro_init_methods(text_widget_mdef);

	if (!init) {
		init = 1;
		if(Py_AtExit(pyro_cleanup) != 0)
			fprintf(stderr,
			  "rivet: warning: cleanup procedure not registered\n");
	}

	if(PyErr_Occurred())
		Py_FatalError("Can not initialize module rivet.");
}
 

