# utility to position objects in a table

import rivet

class table:
	def __init__(self, parent, *configops1, **configops2):
		self.__widget = rivet.canvas(parent, configops1, configops2)

		# padding is necessary if the widget is a canvas, because
		# border offsets need to be included when placing children
		# inside canvases.  Note however that widths and heights
		# set by configure includes widths and heights for border.
		# Both these rules are just the opposite if the widget is
		# a frame.  This code assumes we are using a canvas, because
		# drawlines will not work with frames.

		#
		# All of these should be assumed to be public:
		# 
		self.nrows = 1
		self.ncolumns = 1
		# gaps are on both sides of a cell, lines are draw between gaps
		self.gapw = 1
		self.gaph = 1
		self.defaultanchor = 'c'
		self.drawcolumnlines = 1
		self.drawrowlines = 1
		self.drawborderlines = 1
		self.linecolor = None

		#
		# This may be externally modifyable, but carefully.
		#
		self.array = {}
		self.colors = {}

		#
		# List of objects that span more than one row or column
		#
		self.spanninglist = []

	__getattr__ = rivet.__getattr__

	def additem(self, item, anchor, row, column, rows=1, columns=1):
		if rows < 1: rows = 1
		if columns < 1: columns = 1

		# grow automatically as objects are added
		self.nrows = max(self.nrows, row+rows)
		self.ncolumns = max(self.ncolumns, column+columns)

		self.array[(row, column)] = item
		item._tableanchor = anchor

		# Important to place the widgets, eventhough their container
		# is not packed yet, so that their sizes will be valid.
		# Put them offscreen for now.
		rivet.place(item, '-x', -2000, '-y', -2000)

		# item spans rows or columns, for these only the northwest
		# index points to the item.  Things that span cells are
		# identifyable through their '_tablecellrange' attribute
		if rows > 1 or columns > 1:
			rows = max(rows, 1)
			columns = max(columns, 1)
			# stash away for later use
			item._tablecellrange = (rows, columns)

			# clean out possibly previously occupied table locations
			for r in range(row, row+rows):
				for c in range(column, column+columns):
					if r == row and c == column:
						continue
					try:	del self.array[(r, c)]
					except:	pass

	def delitem(self, row, col):
		# There may be more cleanup than this.
		del self.array[(row, col)]

	def setcolor(self, color, row, column, rows=1, columns=1):
		self.colors[(row, column)] = (color, rows, columns)

	# layout routines...

	def _span_sizes(self, maxs, gap, dimension):

		# this function handles one dimension per call, adjusting
		# the size of the columns or rows in case more space is
		# needed by a cell that spans multiple columns or rows

		for position in self.spanninglist:
			item = self.array[position]

			# start is index of start of span
			# span is length of span
			# size is required size of span
			if dimension == 'row':
				start = position[0]
				span = item._tablecellrange[0]
				size = rivet.winfo.height(item)
			else:
				start = position[1]
				span = item._tablecellrange[1]
				size = rivet.winfo.width(item)

			availsize = 0
			for index in range(start, start+span):
				availsize = availsize + maxs[index]
 
			if availsize + (gap+1+gap) * (span-1) >= size:
				continue

			# we need more space to fit multicell object
			# than has been provided.  Add some space to each.

			# total amount needed is...
			needed = size - availsize - (gap+1+gap) * (span-1)

			# delta is how much to add to each cell.
			# mod is remainder, which should be spread among
			# cells.
			delta, mod = divmod(needed, span)

			for index in range(start, start+span):
				newsize = maxs[index]+delta
				if mod > 0:
					newsize = newsize+1
					mod = mod-1
				maxs[index] = newsize
		return maxs


	def _get_row_heights(self):
		maxrowh = {}
		multicell = []
		array = self.array
		height = rivet.winfo.height
		for row in range(self.nrows):
			hmax = 0
			for col in range(self.ncolumns):
				try:
					item = array[(row, col)]
				except KeyError:
					continue

				try:
					d = item.__dict__
				except:
					# proactively remove these
					# uncooperative objects
					del array[(row, col)]
					continue

				if d.has_key('_tablecellrange'):
					# spanning, handle later
					multicell.append( (row, col) )
					continue

				hmax=max(hmax, height(item))
			maxrowh[row] = hmax
		self.spanninglist = multicell
		if len(self.spanninglist) > 0:
			maxrowh = self._span_sizes(maxrowh, self.gaph, 'row')
		return maxrowh


	def _get_col_widths(self):
		maxcolw = {}
		array = self.array
		width = rivet.winfo.width
		for col in range(self.ncolumns):
			wmax = 0
			for row in range(self.nrows):
				try:
					item = array[(row, col)]
				except KeyError:
					continue

				try:
					d = item.__dict__
				except:
					# proactively remove these
					# uncooperative rivetobjs
					del array[(row, col)]
					continue

				if d.has_key('_tablecellrange'):
					# spanning, handle later
					continue

				wmax=max(wmax, width(item))
			maxcolw[col] = wmax
		if len(self.spanninglist) > 0:
			maxrowh = self._span_sizes(maxcolw, self.gapw, 'column')
		return maxcolw


	def _place_item(self, item, cellx, celly, cellwidth, cellheight):
		try:
			anchor = item._tableanchor
		except AttributeError:
			anchor = self.defaultanchor
		if anchor == None:
			anchor = self.defaultanchor


		width = rivet.winfo.width(item)	
		height= rivet.winfo.height(item)	

		if 'n' in anchor:
			ypos = celly
		elif 's' in anchor:
			ypos = celly + cellheight - height
		else: # center
			ypos = celly + (cellheight - height) / 2

		if 'w' in anchor:
			xpos = cellx
		elif 'e' in anchor:
			xpos = cellx + cellwidth - width
		else: # center
			xpos = cellx + (cellwidth - width) /2

		rivet.place(item, '-x', xpos, '-y', ypos)

	def _place_span(self, item, row, col, maxrowh, maxcolw, cellx, celly):
		spanr, spanc = item._tablecellrange
		cellheight = 0
		cellwidth = 0
		for r in range(row, row+spanr):
			cellheight = cellheight+maxrowh[r]
		cellheight = cellheight + (self.gaph+1+self.gaph) * (spanr - 1)
		for c in range(col, col+spanc):
			cellwidth = cellwidth+maxcolw[c]
		cellwidth = cellwidth + (self.gapw+1+self.gapw) * (spanc - 1)
		self._place_item(item, cellx, celly, cellwidth, cellheight)

	def _place_all(self):
		maxrowh = self.maxrowh
		maxcolw = self.maxcolw
		# add 2 for focus ring, 1 for line, gap for first gap
		pad = self.__widget.cget('bd') + 2 + 1
		celly = pad + self.gaph
		array = self.array
		for row in range(self.nrows):
		    cellx = pad + self.gapw
		    cellheight = maxrowh[row]
		    for col in range(self.ncolumns):
			cellwidth = maxcolw[col]
			try:
				item = array[(row, col)]
				if item.__dict__.has_key('_tablecellrange'):
					self._place_span(item,
						row, col,
						maxrowh, maxcolw,
						cellx, celly)
				else:
					self._place_item(item,
						cellx, celly,
						cellwidth, cellheight)
			except (KeyError, rivet.error):
				pass
			# add 2 gaps, plus one pixel for line
			cellx = cellx + cellwidth + self.gapw + 1 + self.gapw
		    celly = celly + cellheight + self.gaph + 1 + self.gaph


	def arrange(self):
		# place all the widgets as they should be
		rivet.update()	# make sure that sizes have been set

		# 1: get all the max value widths and heights.
		# get_row_heights must be called first because it
		# recognizes items that span cells.
		self.maxrowh = self._get_row_heights()
		self.maxcolw = self._get_col_widths()

		# 2: move the widgets about
		self._place_all()

		# 3: set the frame size accordingly
		# calculate internal sizes.
		totalw = 0
		for i in self.maxcolw.values(): totalw = totalw + i
		totalh = 0
		for i in self.maxrowh.values(): totalh = totalh + i
		# don't need to include `bd' in calculations of canvas size.
		# multiply gap size by 2 to include gaps on both sides
		# of cell, add 1 for each line drawn.
		self.__widget.configure(
		    'width',	totalw +
				self.gapw * self.ncolumns * 2 +
				self.ncolumns + 1,
		    'height',	totalh +
				self.gaph * self.nrows * 2 +
				self.nrows + 1)

		self._drawlines()



	# other routines

	def _spanned_cells(self):
		# This function returns an array with all the positions
		# occupied by spanning items set to that item
		# this is different than how it is stored in self.array,
		# where there is only one entry in the array for a
		# spanning object.
		# non spanning objects are not included.
		rv = {}
		for position in self.spanninglist:
			row = position[0]
			col = position[1]
			item = self.array[position]
			nrows = item._tablecellrange[0]
			ncols = item._tablecellrange[1]
			for r in range(row, row+nrows):
				for c in range(col, col+ncols):
					rv[(r,c)] = item
		return rv


	def _drawlines(self):
		# clean out old lines:
		self.__widget.delete('all')
 		# collinepos and rowlinepos are arrays containing the 
		# pixel locations of the center of the borders between items
		# plus sentinels on the begin and end for 0 and width or height.

		if not self.drawcolumnlines and \
		   not self.drawrowlines and 	\
		   not self.drawborderlines:
			return

		if self.linecolor:
			fglineopt = ('-fill', self.linecolor)
		else:
			fglineopt = ()

		# 2 is added here for focus ring
		origin = self.__widget.cget('bd') + 2

		collinepos = []
		sum = origin
		for col in range(self.ncolumns):
			gapoffset = (col) + (col*self.gapw*2)
			collinepos.append(sum + gapoffset)
			sum = sum + self.maxcolw[col]
		# last item is location beyond last column
		collinepos.append(sum +
			(self.ncolumns) + (self.ncolumns * self.gapw *2))

		rowlinepos = []
		sum = origin
		for row in range(self.nrows):
			gapoffset = (row) + (row*self.gaph*2)
			rowlinepos.append(sum + gapoffset)
			sum = sum + self.maxrowh[row]
		# last item is location beyond last row
		rowlinepos.append(sum +
			(self.nrows) + (self.nrows * self.gaph *2))
		yend = rowlinepos[self.nrows]
		xend = collinepos[self.ncolumns]

		# draw complete lines...
		if self.drawcolumnlines:
			for coll in collinepos:
				self.__widget.create('line',
					coll, 0,
					coll, yend,
					fglineopt)

		if self.drawrowlines:
			for rowl in rowlinepos:
				self.__widget.create('line',
					0, rowl,
					xend, rowl,
					fglineopt)

		# draw background colors
		for item in self.colors.items():
			row = item[0][0]
			col = item[0][1]
			color = item[1][0]
			spanr = item[1][1]
			spanc = item[1][2]

			self.__widget.create('rectangle',
				collinepos[col], rowlinepos[row],
				collinepos[col+spanc], rowlinepos[row+spanr],
				'-fill', color)

		# "erase" places where spanned cells cover lines
		eraseopt = ('-fill', self.__widget.cget('bg'))
		for position in self.spanninglist:
			if self.colors.has_key(position):
				# it's already covered by filled rectangle
				continue

			row = position[0]
			col = position[1]
			item = self.array[position]
			nrows = item._tablecellrange[0]
			ncols = item._tablecellrange[1]

			if self.drawcolumnlines:
			  for r in range(row, row+nrows):
			    for c in range(col+1, col+ncols):
				self.__widget.create('line',
					collinepos[c], rowlinepos[r]+1,
					collinepos[c], rowlinepos[r+1],
					eraseopt)

			if self.drawrowlines:
			  for c in range(col, col+ncols):
			    for r in range(row+1, row+nrows):
				self.__widget.create('line',
					collinepos[c]+1, rowlinepos[r],
					collinepos[c+1], rowlinepos[r],
					eraseopt)

		if self.drawborderlines:
			self.__widget.create('line',
				origin, origin,
				origin, yend,
				xend, yend,
				xend, origin,
				origin, origin,
				fglineopt)


	def oldcode(self):
		# this code draws lines for each gap between each cell.
		# it turned out to be too slow, so the above "subtractive"
		# solution was implemented.  The above solution assumes that
		# spanning will be less frequent than non-spanning.

		spanned = self._spanned_cells()

		# test
		# spanned = {(1,1):1, (2,1):1, (1,2):2, (2,2):2}

		# draw columns
		if self.drawcolumnlines:
		    for row in range(self.nrows):
			ystart = rowlinepos[row]
			yend = rowlinepos[row+1]
			for col in range(self.ncolumns-1):

				try:	item1 = spanned[(row, col)]
				except: item1 = 'spam'
				try:	item2 = spanned[(row, col+1)]
				except: item2 = 'eggs'

				if item1 == item2:
					continue
			
				coll = collinepos[col+1]
				self.__widget.create('line',
					coll, ystart, coll, yend,
					lineopt)

		# draw rows
		if self.drawrowlines:
		  for col in range(self.ncolumns):
			xstart = collinepos[col]
			xend = collinepos[col+1]
			for row in range(self.nrows-1):

				try:	item1 = spanned[(row, col)]
				except: item1 = 'spam'
				try:	item2 = spanned[(row+1, col)]
				except: item2 = 'eggs'

				if item1 == item2:
					continue
			
				rowl = rowlinepos[row+1]
				self.__widget.create('line',
					xstart, rowl, xend, rowl,
					lineopt)


	def _debug_print(self):
		l = self.array.keys()
		l.sort()
		print l

def tablelayoutdemo():
	# example
	r = rivet.init()

	t = table(r)
	t1 = table(t, bd=4, relief='raised')
	t2 = table(t)
	t3 = table(t)
	t4 = table(t)
	t5 = table(t)

	options = ('width', 60, 'bd', 3, 'relief', 'raised')

	t.additem(rivet.label(t, 'text', 'Anchors...', options),
			None, 0, 0)
	t.additem(t1,	None, 1, 0)

	t.additem(rivet.label(t, 'text', 'Span horizontal...', options),
			None, 2, 0)
	t.additem(t2,	None, 3, 0)

	t.additem(rivet.label(t, 'text', 'Span vertical...', options),
			None, 4, 0)
	t.additem(t3,	None, 5, 0)

	t.additem(rivet.label(t, 'text', 'Sparse...', options),
			None, 6, 0)
	t.additem(t4,	None, 7, 0)

	t.additem(rivet.label(t, 'text',
			'Span row & column, and using some color', options),
			None, 8, 0)
	t.additem(t5,	None, 9, 0)

	t1.additem(rivet.label(t1, 'text', 'nw'), 'nw', 0, 0)
	t1.additem(rivet.label(t1, 'text', 'n'),  'n',  0, 1)
	t1.additem(rivet.label(t1, 'text', 'ne'), 'ne', 0, 2)

	t1.additem(rivet.label(t1, 'text', 'w'),  'w', 1, 0)
	t1.additem(rivet.label(t1, 'text', 'center'), '', 1, 1)
	t1.additem(rivet.label(t1, 'text', 'e'),  'e', 1, 2)

	t1.additem(rivet.label(t1, 'text', 'sw'), 'sw', 2, 0)
	t1.additem(rivet.label(t1, 'text', 's'),  's',  2, 1)
	t1.additem(rivet.label(t1, 'text', 'se'), 'se', 2, 2)

	# demo use of line color
	t1.linecolor = 'white'

	t2.additem(rivet.button(t2, 'text', 'columns 0-1', 'width', 30),
			None, 0, 0, 0, 2)
	t2.additem(rivet.label(t2, 'text', 'row 1, col 0'),
			None, 1, 0)
	t2.additem(rivet.label(t2, 'text', 'row 1, col 1'),
			None, 1, 1)
	t2.additem(rivet.button(t2, 'text', 'columns 0-1', 'width', 30),
			None, 2, 0, 0, 2)

	t3.additem(rivet.scale(t3, 'showvalue', 0),
			None, 0, 0, 2, 0)
	t3.additem(rivet.label(t3, 'text', 'row 0, column 1'),
			None, 0, 1)
	t3.additem(rivet.label(t3, 'text', 'row 1, column 1'),
			None, 1, 1)
	t3.additem(rivet.scale(t3, 'showvalue', 0),
			None, 0, 2, 2, 0)

	t4.additem(rivet.label(t4, 'text', 'row 2, col 1'),
			None, 2, 1)
	t4.additem(rivet.label(t4, 'text', 'row 1-3, col 4'),
			None, 1,4,3,1)
	t4.additem(rivet.label(t4, 'text', 'row 5, col 1-2'),
			None, 5,1,1,2)

	t5.additem(rivet.label(t5, 'text', 'row 0, col 0'),
			None, 0,0)
	t5.additem(rivet.label(t5, 'text', 'row 1, col 0'),
			None, 1,0)
	t5.additem(rivet.label(t5,
			'text', 'row 0-2, col 1-2',
			'bg', 'white'),
		None, 0,1,3,2)
	t5.setcolor('white', 0,1,3,2)
	t5.additem(rivet.label(t5,
			'text', 'row 4-5, col 0-1',
			'bg', 'yellow'),
		None, 4,0,2,2)
	t5.setcolor('yellow', 4,0,2,2)
	t5.additem(rivet.label(t5, 'text', 'row 0-3, col 3'),
			None, 0,3,4,1)
	t5.additem(rivet.label(t5, 'text', 'row 4, col 2-3'),
			None, 4,2,1,2)
	t5.gapw = 15
	t5.gaph = 0.5

	t1.arrange()
	t2.arrange()
	t3.arrange()
	t4.arrange()
	t5.arrange()

	t.drawborderlines = 0
	t.drawrowlines = 0
	t.drawcolumnlines = 0
	t.arrange()
	t.pack()
	t.configure(bg='red')

	rivet.mainloop()


if __name__ == '__main__':
	tablelayoutdemo()

