#include <rivet.h>

/* This code was adapted from tkEvent.c */


/*
 * One of the following structures exists for each file with a handler
 * created by the "fileevent" command.  Several of the fields are
 * two-element arrays, in which the first element is used for read
 * events and the second for write events.
 */

typedef struct FileEvent {
    int fd;				/* handle for file. */
    Callback callbacks[2];		/* Scripts to evaluate in response to
					 * events (malloc'ed).  NULL means no
					 * handler for event. */
    struct FileEvent *nextPtr;		/* Next in list of all file events
					 * currently defined. */
} FileEvent;

static FileEvent *firstFileEventPtr = NULL;
static int	rivet_fileeventproc _ANSI_ARGS_((ClientData clientData,
				int mask, int flags));

/* ARGSUSED */
Callback
rivet_get_fileevent(fd, mode)
	int fd;
	int mode; /* 0 for readable, 1 for writeable */
{
	int index;
	FileEvent *fevPtr;
	Callback cb;

	if(mode != 0 && mode != 1)
		return 0; /* caller error */
	index = mode;

	/*
	 * Locate an existing file handler for this file, if one exists,
	 * and make a new one if none currently exists.
	 */

	for (fevPtr = firstFileEventPtr; fevPtr; fevPtr = fevPtr->nextPtr) {
		if (fevPtr->fd == fd) {
			cb = fevPtr->callbacks[index];
			rivet_incref_callback(cb);
			return cb;
		}
	}
	return 0;
}


/* ARGSUSED */
void
rivet_set_fileevent(fd, mode, cb)
	int fd;
	int mode; /* 0 for readable, 1 for writeable */
	Callback cb;
{
	int index;
	FileEvent *fevPtr, *prevPtr;

	if(mode != 0 && mode != 1)
		return ; /* caller error */
	index = mode;

	/*
	 * Locate an existing file handler for this file, if one exists,
	 * and make a new one if none currently exists.
	 */

	for (fevPtr = firstFileEventPtr; ; fevPtr = fevPtr->nextPtr) {
		if (fevPtr == NULL) {
			fevPtr = (FileEvent *) ckalloc(sizeof(FileEvent));
			fevPtr->fd = fd;
			fevPtr->callbacks[0] = NULL;
			fevPtr->callbacks[1] = NULL;
			fevPtr->nextPtr = firstFileEventPtr;
			firstFileEventPtr = fevPtr;
			Tk_CreateFileHandler2(fd, rivet_fileeventproc,
					(ClientData) fevPtr);
			break;
		}
		if (fevPtr->fd == fd) {
			break;
		}
	}

	/*
	 * If we're supposed to delete the event handler, do so.
	 */
	if (rivet_is_null_callback(cb)) {
		if (fevPtr->callbacks[index] != NULL) {
			rivet_decref_callback(fevPtr->callbacks[index]);
			ckfree(fevPtr->callbacks[index]);
			fevPtr->callbacks[index] = NULL;
		}
		if ((!fevPtr->callbacks[0]) && (!fevPtr->callbacks[1])) {
			if (firstFileEventPtr == fevPtr) {
				firstFileEventPtr = fevPtr->nextPtr;
			} else {
				for (prevPtr = firstFileEventPtr;
					prevPtr->nextPtr != fevPtr;
					prevPtr = prevPtr->nextPtr) {
					/* Empty loop body. */
				}
				prevPtr->nextPtr = fevPtr->nextPtr;
			}
			Tk_DeleteFileHandler(fevPtr->fd);
			ckfree((char *) fevPtr);
		}
		return ;
	}

	/*
	 * This is a new handler being created.  Save its script.
	 */
	if (fevPtr->callbacks[index] != NULL) {
		rivet_decref_callback(fevPtr->callbacks[index]);
		ckfree(fevPtr->callbacks[index]);
	}
	rivet_incref_callback(cb);
	fevPtr->callbacks[index] = cb;
	return ;
}

/*
 *----------------------------------------------------------------------
 *
 * rivet_fileeventproc --
 *
 *	This procedure is invoked by Tk's event loop to deal with file
 *	event bindings created by the "fileevent" command.
 *
 * Results:
 *	The return value is TK_FILE_HANDLED if the file was ready and
 *	a script was invoked to handle it.  Otherwise an OR-ed combination
 *	of TK_READABLE and TK_WRITABLE is returned, indicating the events
 *	that should be checked in future calls to select.
 *
 * Side effects:
 *	Whatever the event script does.
 *
 *----------------------------------------------------------------------
 */
static int
rivet_fileeventproc(clientData, mask, flags)
    ClientData clientData;	/* Pointer to FileEvent structure for file. */
    int mask;			/* OR-ed combination of the bits TK_READABLE,
				 * TK_WRITABLE, and TK_EXCEPTION, indicating
				 * current state of file. */
    int flags;			/* Flag bits passed to Tk_DoOneEvent;
				 * contains bits such as TK_DONT_WAIT,
				 * TK_X_EVENTS, Tk_FILE_EVENTS, etc. */
{
    FileEvent *fevPtr = (FileEvent *) clientData;
    int fd;
    int code, checkMask;

    if (!(flags & TK_FILE_EVENTS)) {
	return 0;
    }

    /*
     * The code here is a little tricky, because the script for an
     * event could delete the event handler.  Thus, after we call
     * Tcl_GlobalEval we can't use fevPtr anymore.  We also have to
     * copy the script to make sure that it doesn't get freed while
     * being evaluated.
     */

    checkMask = 0;
    fd = fevPtr->fd;
    if (fevPtr->callbacks[1] != NULL) {
	if (mask & TK_WRITABLE) {
		code = rivet_call_callback(
			fevPtr->callbacks[1], 0, 0, 0, RIVET_CB_FILEEVENT);
		if (code != TCL_OK) {
			goto error;
		}
		return TK_FILE_HANDLED;
	} else {
		checkMask |= TK_WRITABLE;
	}
    }
    if (fevPtr->callbacks[0] != NULL) {
	if ((mask & TK_READABLE)) { /* gack, problem with stdio buffering. */
		code = rivet_call_callback(
			fevPtr->callbacks[0], 0, 0, 0, RIVET_CB_FILEEVENT);
	    	if (code != TCL_OK) {
			goto error;
		}
		return TK_FILE_HANDLED;
	} else {
		checkMask |= TK_READABLE;
	}
    }
    return checkMask;

    /*
     * An error occurred in the script, so we have to call
     * Tk_BackgroundError.  However, it's possible that the file ready
     * condition didn't get cleared for the file, so we could end
     * up in an infinite loop if we're not careful.  To be safe,
     * delete the event handler.
     */

    error:
    rivet_set_fileevent(fd, (mask&TK_WRITABLE)?1:0, (Callback)"");
    Tcl_AddErrorInfo(rivet_main_window->interp,
	    "\n    (script bound to file event - binding deleted)");
    Tk_BackgroundError(rivet_main_window->interp);
    return TK_FILE_HANDLED;
}

